// -----------------------------------------------------------------------------
// Template.java
// -----------------------------------------------------------------------------

/*
 * =============================================================================
 * Applet that simulates the interaction of bacteria and phage on a lattice
 * By Jan O. Haerter
 * =============================================================================
 */



import java.io.*;
import java.awt.*;
import java.awt.event.*;
import java.net.*;
import java.util.*;
import java.util.Hashtable;
import java.util.Vector;
import java.util.Random;
import java.util.Date;
import javax.swing.*;



// change to previous version: caseNumber5 with co-existence of phage of different g_p
public class CRISPR_evol_ph_evol extends Panel implements ActionListener, AdjustmentListener, Runnable, MouseListener {

    private static final long serialVersionUID = 1L;
    
    //crucial parameters

    public String runNumber="testing_kim_"; // "del_hi_conv1_rprime_test_28_10_2011_CRISPR_random_";

    public double expMin =0.37;
    public double expMax =1.0;
    public double expInc =0.025;

    public double wTMin  =1.;
    public double wTMax  =1.;
    public double wTInc  =0.25;

    public double rbMin  =2.;//10//1.301;//1.301;//1.602;//1.602;//1.301
    public double rbMax  =2.;//10.;//1.301;//1.301;//1.602;//1.602;//1.301
    public double rbInc  =0.5;

    public double NPMin  =2.0;
    public double NPMax  =10.0;
    public double NPInc  =2.0;

    public double rPrime = 0.0001;//0.000000001;//0.001;//0.00001;
    public double cost   = 1.6;//0.125/2.;//1/20.;

    public double immSwitchRate = Math.pow(10,-rbMin);

    public boolean randomUpdate=false;//false;//should random updating of sites be performed?
    public boolean bitRandomUpdate=false;
    public double jumpPhage=2.;
    public double jumpBact=0.;
    public int caseNumber=9;
    public int NPHAGE=1; // for the situation of multiple phage species (case 4)
    public int NRES=1;
    public double convCrit=.00001;//0.00001;//0.0000001;//0.00001;//0.001//0.00025//0.000025
    public double convCritSlope=.00001;//0.00001;//0.0000001;//0.000005;//0.00001;//0.0001;//0.000025
    public boolean expevolve = false;
    public boolean wTevolve  = true; // in case 3: should phage evolve also in their velocity?
    public boolean addExtraBacteria = false;
    public double RM=1.0;
    public int nResFrac=3;
    public boolean countExp=true;
    public boolean zeroInitRes=false;


    //Constants: Dynamics
    public static int L =150; //Size of lattice
    public static int Nneighbors = 4;
    public static int NBACT = 1;
    public int2D[] connections;
    public double[] waitingTimes=new double[4];
    public double initialTime=0.0;
    public boolean addphage, addphage1;
    public int ctime=0;
    public int maxBacteriaSpecies=20;
    public int maxPhageSpecies=360;
    public double expressionAa=expMin;
    //    public boolean insertPhage_a=true;
    //public boolean insertPhage_b=false;
    public boolean phageAdded=false;
    public boolean bacteriaAdded=false;
    public boolean bacteriaAdded2=false;
    public PrintWriter pw;
    

    //    HostMute[] randomH=new HostMute[Nneighbors];

    int delay = 2; 
    int maxImmunity = 2000;
    public int updateEveryM0 = L * L / 100;
    public int updateEveryM = 2000*updateEveryM0; //How many updates before lattice is redrawn
    //    public int diseasePool = 10 * L;
    double alpha = 0.05 / (L * L); //how often new diseases appear
	
    //Constants: Heights, widths and sizes
    public static int nodeSize = 2; //length and height of one node
    public static int PlotWidth = 200, PlotHeight = 100; //height and width of diversity plot
    public static int x0 = 20, y0 = 20; //starting coordinates for lattice
    public static int panelX = x0 + L*nodeSize + 50, panelY = 40; //x and y coordinates for panel
    static final int appletHeight = L*nodeSize+2*x0;
    static final int appletWidth = L*nodeSize + 200; //height and width of Applet
        
    //Initializing parameters
    boolean keepMoving = true;
    boolean periodicBC = true;
    public HostMute[] host;
    //public Host[] host;
    URL codeBase;
    String standardFilename;

    outputWriter oW;
    outputWriter oW_hist,oW_hist1;
    outputWriter oW_scatter,oW_scatter_bact,oW_scatter_phage;
    outputWriter oW_interaction;
    outputWriter oW_attackSurvival;

    //currentPosition etc.
    int currentX,currentY;
    double currentT;
    int currentI;
    //number of bacteria and phage on lattice
    double[] nBact=new double[maxBacteriaSpecies];
    double[] nPhag=new double[maxPhageSpecies];
    double[] nPhagExp=new double[maxPhageSpecies];
    double[] nPhagwT=new double[maxPhageSpecies];

    makeAverage[] nBactTot   =new makeAverage[maxBacteriaSpecies];
    makeAverage[] nPhagTot   =new makeAverage[maxPhageSpecies];
    makeAverage[] nPhagExpTot=new makeAverage[maxPhageSpecies];
    makeAverage[] nPhagwTTot =new makeAverage[maxPhageSpecies];

    makeAverage[] phagExp =new makeAverage[maxPhageSpecies];
    makeAverage[] phagAdd =new makeAverage[maxPhageSpecies];

    
    LeastSquares[] nBactTotLS  =new LeastSquares[maxBacteriaSpecies];
    LeastSquares[] nPhagTotLS  =new LeastSquares[maxPhageSpecies];
    LeastSquares[] nPhagExpTotLS  =new LeastSquares[maxPhageSpecies];
    LeastSquares[] nPhagwTTotLS  =new LeastSquares[maxPhageSpecies];

    

    double countAve=0.;
    int nPars;
    double[] pars;
    double[] parsMin;
    double[] parsMax;
    double[] parsInc;

    int [][] attackSurvival = new int[NPHAGE+1][7]; // 0=attack, 1=survival
    
	
    int countT=0;
    long ti0=0;
    long ti1=0;
    
    

    //graphics
    private volatile Thread animator = null;
    Image background2D;
    Graphics2D bufferedGraphics2D;
    RenderingHints hints;
    BasicStroke line_stroke;
    //    Colormap cmapTW;
    
    GradientPaint redtowhite, redtowhite1;
    Color lblue, white, yellow, lyellow, lbluish, navyblue, black, golden, nColor[], beige, orange;
    Color dgreen, lgreen, red, lred, dgrey, green, mblack, lgrey;
    Color cbackground, cborder;
    Color ctext, cbuttontext;
    
    //Main buttons
    Button button_stop,button_print;
    Button button_periodicBC;
    
    //Scrollbars
    Scrollbar scroll_alpha, scroll_genTime1, scroll_genTime2;
    
    double alphaN = alpha*(L*L), alphaNMin = 0.001, alphaNMax = 10;
    int nBINS = 100;
    int alphaX, alphaY;
    int gen1X, gen1Y, gen2X, gen2Y;
    Scrollbar scroll_time;

    // vectors to hold a collection of the introduced phage and bacteria
    Vector<Integer>    phageCollection    = new Vector<Integer>();
    Vector<Integer> bacteriaCollection    = new Vector<Integer>();

    Vector<Double> fitx=new Vector<Double>();
    Vector<Double> fity=new Vector<Double>();

    /** The tree root. */
    protected BinaryNode root;
    


	BufferedWriter out;
	// set up a file to write output to
    /*
 
    */

    public static void main(String[] args) {
	Frame f = new Frame();
	f.addWindowListener(new java.awt.event.WindowAdapter() {
		public void windowClosing(java.awt.event.WindowEvent e) {
		    System.exit(0);
		};
	    });

	CRISPR_evol templ = new CRISPR_evol();
	int sizeX, sizeY;
	
	sizeX=nodeSize*L+250;
	sizeY=nodeSize*L+120;
	templ.setSize(sizeX,sizeY); // same size as defined in the HTML APPLET
	f.add(templ);
	f.pack();
	templ.init();
	f.setSize(sizeX,sizeY); // add 20, seems enough for the Frame title,
	//	f.show();
	f.setVisible(true);
    }

    /**
     * Initializes the applet by invoking createGUI
     */
  
	public void init() {
        try {
            SwingUtilities.invokeAndWait(new Runnable() {
		    public void run() { 
			//codeBase = getCodeBase();
			createGUI(); 
		    }
		});
        } catch (Exception e) {
            System.err.println("createGUI didn't successfully complete");
        }

	/*
	
	
	try {
	    out = new BufferedWriter(new FileWriter("outfilename"));
	    out.write("aString");
	    out.close();
	} catch (IOException e) {
	}
	*/
    }
    
    /**
     * makes graphical user interface, made by Ala
     */
    public void createGUI() {
        setSize(appletWidth, appletHeight);
	
        background2D = this.createImage(appletWidth, appletHeight);
        bufferedGraphics2D = (Graphics2D) background2D.getGraphics();
        start();

	/*
	try{
	    pw = new PrintWriter(new FileWriter("test.dat"));
	    //this is equal to:
	    //pw = new PrintWriter(new FileWriter(filename, false));
	    
	}
    
	catch (IOException e) {
	    System.out.println("couldn't open");
	    e.printStackTrace();
	}
	*/


        //Define colours
        white = new Color(255, 255, 255);
        lred = new Color(255, 153, 0);
        lblue = new Color(190, 203, 255);
        red = new Color(255, 42, 42);
        navyblue = new Color(0, 81, 186);
        black = new Color(0, 0, 0);
        orange = new Color(232, 170, 63);
        dgrey = new Color(51, 51, 51);
        lgrey = new Color(150, 150, 150);
        lgreen = new Color(0, 128, 0);
        cbackground = white;

        line_stroke = new BasicStroke(2.0f, BasicStroke.CAP_BUTT, BasicStroke.JOIN_BEVEL); //???

        hints = new RenderingHints(RenderingHints.KEY_ALPHA_INTERPOLATION, RenderingHints.VALUE_ALPHA_INTERPOLATION_SPEED);
        hints.put(RenderingHints.KEY_ANTIALIASING, RenderingHints.VALUE_ANTIALIAS_ON);
        hints.put(RenderingHints.KEY_COLOR_RENDERING, RenderingHints.VALUE_COLOR_RENDER_SPEED);
        hints.put(RenderingHints.KEY_DITHERING, RenderingHints.VALUE_DITHER_DISABLE);
        hints.put(RenderingHints.KEY_FRACTIONALMETRICS, RenderingHints.VALUE_FRACTIONALMETRICS_ON);
        hints.put(RenderingHints.KEY_ANTIALIASING, RenderingHints.VALUE_ANTIALIAS_ON);
        hints.put(RenderingHints.KEY_RENDERING, RenderingHints.VALUE_RENDER_SPEED);
        hints.put(RenderingHints.KEY_STROKE_CONTROL, RenderingHints.VALUE_STROKE_NORMALIZE);
        hints.put(RenderingHints.KEY_TEXT_ANTIALIASING, RenderingHints.VALUE_TEXT_ANTIALIAS_OFF);


        //Buttons
        alphaX = panelX + 40;
        alphaY = panelY-4;
	gen1X  = panelX + 160;
	gen1Y  = panelY+90-4;
	gen2X  = panelX + 160;
	gen2Y  = panelY+135-4;
        //Button parameters
        int butWidth = 60;
        int butHeight = 25;
       

	bufferedGraphics2D.setFont(new Font("SansSerif", Font.BOLD, 10));


        // Add scroll bar Update Speed
        panelY += 45;
        bufferedGraphics2D.drawString("Update Speed", panelX, panelY-4);
        scroll_time = new Scrollbar(Scrollbar.HORIZONTAL, 10, 3, 1, 103);
        scroll_time.addAdjustmentListener(this);
        scroll_time.setBounds(panelX, panelY, 100, 15);
        scroll_time.setBackground(Color.white);
        add(scroll_time);
        scroll_time.setValue(updateEveryM/updateEveryM0);
        updateEveryM = updateEveryM0 * scroll_time.getValue();
               
        // Add scroll bar Gen Times, scrollbar for generation time of bacteria species 1 relative to phage reproduction time.
	panelY += 45;
       // bufferedGraphics2D.setFont(new Font("SansSerif", Font.BOLD, 10));
	if (NBACT>0) {
	    double minval=0.02;
	    double maxval=50.0;
	    //	    bufferedGraphics2D.setColor(Color.red);
	     bufferedGraphics2D.drawString("rel. bacteria 1 gen. time", panelX , panelY-4);
	    bufferedGraphics2D.drawString(""+minval, panelX-25 , panelY+11);
	    bufferedGraphics2D.drawString(""+maxval, panelX+105 , panelY+11);

	    
	    scroll_genTime1 = new Scrollbar(Scrollbar.HORIZONTAL, (int)(minval*100.), 50, 2, (int)(maxval*100.+25.));
	    scroll_genTime1.addAdjustmentListener(this);
	    scroll_genTime1.setBounds(panelX, panelY, 100, 15);
	    scroll_genTime1.setBackground(Color.white);
	    
	    add(scroll_genTime1);
	    
	    int genTime  = (int)(100.*wTMin);
	    scroll_genTime1.setValue(genTime);	
	    updateGenTimes(bufferedGraphics2D, gen1X, gen1Y,(double)genTime);
	}

	if (NBACT>1){ 
	    // Add scroll bar Gen Times, this is the scrollbar for generation time of bacteria species 2 relative to phage reproduction
	    panelY += 45;    
	    bufferedGraphics2D.drawString("rel. bacteria 2 gen. time", panelX , panelY-4);
	    
	    scroll_genTime2 = new Scrollbar(Scrollbar.HORIZONTAL, 1, 50, 2, 5025);
	    scroll_genTime2.addAdjustmentListener(this);
	    scroll_genTime2.setBounds(panelX, panelY, 100, 15);
	    scroll_genTime2.setBackground(Color.white);
	    
	    add(scroll_genTime2);
	    
	    int genTime  = (int)(100.*waitingTimes[1]);
	    scroll_genTime2.setValue(genTime);	
	    updateGenTimes(bufferedGraphics2D, gen2X, gen2Y,(double)genTime);
	}
	

        //Add stop button
        panelY += 45;
        button_stop = new Button("Pause1");
        button_stop.setBounds(panelX, panelY, butWidth, butHeight);
        button_stop.setBackground(orange);
        add(button_stop);
        button_stop.addActionListener(this);

        panelY += 4*butHeight;


        panelY += 45;
        button_print = new Button("Print");
        button_print.setBounds(panelX, panelY, butWidth, butHeight);
        button_print.setBackground(orange);
        add(button_print);
        button_print.addActionListener(this);

        panelY += 4*butHeight;

        setLayout(null);
        addMouseListener(this);        
	initOutput();
	initCounter(); // initialize the count of bacteria and phage populations
        initLattice(bufferedGraphics2D);
	initialCondition();	
        bufferedGraphics2D.setBackground(black);
    }
    

    public void initOutput(){


	if (caseNumber==1) {
	    NPHAGE=1;
	    standardFilename="L"+L+"_mix_"+randomUpdate+"_1b1p.dat";
	    nPars=2;
	    pars=new double[nPars];
	    parsMin=new double[nPars];
	    parsMax=new double[nPars];
	    parsInc=new double[nPars];

	    pars[0]    = expMin;
	    pars[1]    = wTMin;
	    parsMin[0] = expMin;
	    parsMin[1] = wTMin;
	    parsMax[0] = expMax;
	    parsMax[1] = wTMax;
	    parsInc[0] = expInc;
	    parsInc[1] = wTInc;
	}	
	if (caseNumber==2) {
	    NPHAGE=2;
	    standardFilename="L"+L+"_mix_"+randomUpdate+"_1b2p.dat";
	    nPars=2;
	    pars=new double[nPars];
	    parsMin=new double[nPars];
	    parsMax=new double[nPars];
	    parsInc=new double[nPars];

	    pars[0]    = expMin;
	    pars[1]    = wTMin;
	    parsMin[0] = expMin;
	    parsMin[1] = wTMin;
	    parsMax[0] = expMax;
	    parsMax[1] = wTMax;
	    parsInc[0] = expInc;
	    parsInc[1] = wTInc;
	}

	if (caseNumber==3) {
	    NPHAGE=1;
	    if (wTevolve && expevolve) {
		nPars=0;
		standardFilename="L"+L+"_mix_"+randomUpdate+"_1b1p_evolPhaAll.dat";
	    }
	    if (!wTevolve && expevolve) {
		nPars=1;
		pars=new double[nPars];		
		parsMin=new double[nPars];
		parsMax=new double[nPars];
		parsInc=new double[nPars];

		pars[0]    = wTMin;
		parsMin[0] = wTMin;
		parsMax[0] = wTMax;
		parsInc[0] = wTInc;		
		standardFilename="L"+L+"_mix_"+randomUpdate+"_1b1p_evolPha_exp.dat";
	    }	 
	    if (wTevolve && !expevolve) {
		nPars=1;
		pars=new double[nPars];		
		parsMin=new double[nPars];
		parsMax=new double[nPars];
		parsInc=new double[nPars];
		
		pars[0]    = expMin;
		parsMin[0] = expMin;
		parsMax[0] = expMax;
		parsInc[0] = expInc;		
		standardFilename="L"+L+"_mix_"+randomUpdate+"_1b1p_evolPha_wT.dat";
	    }
	    if (addExtraBacteria) standardFilename="extraB_"+standardFilename; 
	}
	if (caseNumber==4){
	    nPars=2;
	    pars=new double[nPars];
	    parsMin=new double[nPars];
	    parsMax=new double[nPars];
	    parsInc=new double[nPars];

	    pars[0]    = wTMin;
	    parsMin[0] = wTMin;
	    parsMax[0] = wTMax;
	    parsInc[0] = wTInc;
	    
	    pars[1]    = rbMin;
	    parsMin[1] = rbMin;
	    parsMax[1] = rbMax;
	    parsInc[1] = rbInc;	    

	    standardFilename="L"+L+"_mix_"+randomUpdate+"_1b_evolBac_nphage_"+NPHAGE+"_nres_"+NRES+".dat";
	}

	if (caseNumber==5) {// phages have same exp but different gp
	    NPHAGE=2;
	    standardFilename="L"+L+"_mix_"+randomUpdate+"_1b2p_var_gp.dat";
	    nPars=2;
	    pars=new double[nPars];
	    parsMin=new double[nPars];
	    parsMax=new double[nPars];
	    parsInc=new double[nPars];

	    pars[0]    = expMin;
	    pars[1]    = wTMin;
	    parsMin[0] = expMin;
	    parsMin[1] = wTMin;
	    parsMax[0] = expMax;
	    parsMax[1] = wTMax;
	    parsInc[0] = expInc;
	    parsInc[1] = wTInc;
	}

	if (caseNumber==6){// same as 4 but cycling through the number of added phages
	    nPars=1;
	    pars=new double[nPars];
	    parsMin=new double[nPars];
	    parsMax=new double[nPars];
	    parsInc=new double[nPars];

	    pars[0]    = NPMin;
	    parsMin[0] = NPMin;
	    parsMax[0] = NPMax;
	    parsInc[0] = NPInc;
	    
	    standardFilename="L"+L+"_mix_"+randomUpdate+"_1b_evolBac_nphage_var_max_"+(int)(NPMax)+"_RM_"+RM+".dat";
	}


	if (caseNumber==8){// same as 4 but cycling through the number of added phages
	    nPars=2;
	    pars=new double[nPars];
	    parsMin=new double[nPars];
	    parsMax=new double[nPars];
	    parsInc=new double[nPars];

	    pars[0]    = NPMin;
	    parsMin[0] = NPMin;
	    parsMax[0] = NPMax;
	    parsInc[0] = NPInc;
	    
	    pars[1]    = 1;
	    parsMin[1] = 1;
	    parsMax[1] = pars[0];
	    parsInc[1] = 1;

	    
	    standardFilename="L"+L+"_mix_"+randomUpdate+"_1b_evolBac_nphage_var_max_"+(int)(NPMax)+"_RM_"+RM+".dat";
	}


	if (caseNumber==7){ // same as 4 but with mutating phage, mutation in g_p
	    nPars=1;
	    pars=new double[nPars];
	    parsMin=new double[nPars];
	    parsMax=new double[nPars];
	    parsInc=new double[nPars];
	    
	    pars[0]    = rbMin;
	    parsMin[0] = rbMin;
	    parsMax[0] = rbMax;
	    parsInc[0] = rbInc;	    

	    standardFilename="L"+L+"_mix_"+randomUpdate+"_1b_evolBac_evolPha_nphage_"+NPHAGE+".dat";
	}


	if (caseNumber==9){
	    nPars=2;
	    pars=new double[nPars];
	    parsMin=new double[nPars];
	    parsMax=new double[nPars];
	    parsInc=new double[nPars];

	    pars[0]    = wTMin;
	    parsMin[0] = wTMin;
	    parsMax[0] = wTMax;
	    parsInc[0] = wTInc;
	    
	    pars[1]    = rbMin;
	    parsMin[1] = rbMin;
	    parsMax[1] = rbMax;
	    parsInc[1] = rbInc;	    

	    standardFilename="L"+L+"_mix_"+randomUpdate+"_1b_evolBac_nres_nphage_"+NPHAGE+"_nres_"+NRES+"_immSwRt"+immSwitchRate+"_wTMin_"+wTMin+"_rbMin_"+rbMin+"_rPrime_"+rPrime+"_cost_"+cost+"_zeroInitRes_"+zeroInitRes+"_bitRan_"+bitRandomUpdate+"_jumpP_"+jumpPhage+"_jumpB_"+jumpBact+".dat";
	}


	
	oW                = new outputWriter("std_out_"+runNumber+"_"+standardFilename,true);
	oW_hist           = new outputWriter("hist_nRes_"+runNumber+"_"+standardFilename,true);
	oW_hist1          = new outputWriter("hist_immSwi_"+runNumber+"_"+standardFilename,true);
	oW_scatter        = new outputWriter("scatter_"+runNumber+"_"+standardFilename,true);
	oW_scatter_bact   = new outputWriter("scatter_bact_"+runNumber+"_"+standardFilename,true);
	oW_scatter_phage  = new outputWriter("scatter_phage_"+runNumber+"_"+standardFilename,true);
	oW_interaction    = new outputWriter("interactionMatrix_"+runNumber+"_"+standardFilename,true);
	oW_attackSurvival = new outputWriter("attackSurvival_"+runNumber+"_"+standardFilename,true);

    }

    public void initCounter(){

	for (int i=0;i<NPHAGE+1;i++) for (int j=0;j<attackSurvival[0].length;j++) attackSurvival[i][j]=0;

	for (int b=0; b<maxBacteriaSpecies; b++) {
	    nBactTot[b]=new makeAverage();nBactTot[b].setName("bact"+b);
	    nBactTotLS[b]=new LeastSquares((double)(L*L));
}
	for (int p=0; p<maxPhageSpecies; p++)    {
	    nPhagTot[p]=new makeAverage();nPhagTot[p].setName("phag"+p);
	    nPhagTotLS[p]=new LeastSquares((double)(L*L));
	}
	for (int p=0; p<maxPhageSpecies; p++)    {
	    nPhagExpTot[p]=new makeAverage();nPhagExpTot[p].setName("phagExp"+p);
	    nPhagExpTotLS[p]=new LeastSquares((double)(L*L));
}	
	for (int p=0; p<maxPhageSpecies; p++)    {
	    nPhagwTTot[p] =new makeAverage();nPhagwTTot[p].setName("phagwT"+p);
	    nPhagwTTotLS[p]=new LeastSquares((double)(L*L));
}
	for (int p=0; p<maxPhageSpecies; p++)    {
	    phagExp[p]=new makeAverage();phagExp[p].setName("phagCountExp"+p);	    	    phagAdd[p]=new makeAverage();phagAdd[p].setName("phagCountAdd"+p);	    
	}
	

	

    }

    
    /**
     * Draws lattice with initial infection
     * @param g	Graphics2D
     */
       public void initLattice(Graphics2D g) {
	
	   host = new HostMute[L * L];
	   //host=new Host[L*L];
	   int x = 0, y = 0;
	// good ranges : 1 phage species with imm=1.  [0 .. 2]
	//             : 1 phage species with imm=0.5 [0 .. 15]
	
	waitingTimes[0]=wTMin; // 1st bacteria vs. phage
	waitingTimes[1]=3.0; // 2nd bacteria vs. phage
	waitingTimes[2]=1.; // 1st phage
	waitingTimes[3]=1.; // 2nd phage
	

        if (NBACT>0) {	   
	    scroll_genTime1.setValue((int)(100.*waitingTimes[0]));	
	    updateGenTimes(bufferedGraphics2D, gen1X, gen1Y,waitingTimes[0]);
	}
	if (NBACT>1) {
	    scroll_genTime2.setValue((int)(100.*waitingTimes[1]));	
	    updateGenTimes(bufferedGraphics2D, gen2X, gen2Y,waitingTimes[1]);
	}


	
	addphage =true;
	addphage1=true;

	//	nspecies =0;
	//	nBspecies=bacteriaCollection.size();

	connections=new int2D[Nneighbors];
	connections[0]=new int2D();
	connections[0].init(1,0);
	connections[1]=new int2D();
	connections[1].init(0,1);
	connections[2]=new int2D();
	connections[2].init(-1,0);
	connections[3]=new int2D();
	connections[3].init(0,-1);
	
	

        for (int i = 0; i < L; ++i) {
            for (int j = 0; j < L; ++j) {
		int2D r=index2D(i,j);
		//System.out.println(r.x+ " "+r.y);
		int2D[] neighbors=getNeighbors(r);
	
		x = x0 + nodeSize * j;
                y = y0 + nodeSize * i;

		
		host[i * L + j] = new HostMute(); //create host
		//host[i * L + j] = new Host(); //create host
		
                host[i * L + j].hostID = i * L + j; //name host
                host[i * L + j].init(i, j, maxImmunity, Nneighbors, neighbors); //initialize host
                	     
		
		g.setColor(host[i * L + j].hostColor);
                g.fillRect(x, y, nodeSize, nodeSize);
		
            }
        }

	for (int i = 0; i < L; ++i) {
            for (int j = 0; j < L; ++j) {
		HostMute[] h=new HostMute[Nneighbors];
		//for (int k=0;k<Nneighbors;k++) h[k]=host[index1D(add2D(index2D(i,j),connections[k]))]; this has been updated on 10/18/2011 to include the random update.

		if (!randomUpdate)
		    for (int k=0;k<Nneighbors;k++) h[k]=host[index1D(add2D(index2D(i,j),connections[k]))];
		if (randomUpdate){
		    for (int k=0;k<Nneighbors;k++) h[k]=host[(int)(Math.random()*L*L)];
		}
		host[i*L+j].setNeighbors(h);
	    }
	}

	
       }
    
    // resets the lattice by removing all bacteria and phage from a given site
    public boolean resetLattice(){
	boolean err=false;
	for (int i=0;i<L*L;i++) if (host[i]==null) err=true;
	if (err==false) {
	    for (int i=0;i<L*L;i++) host[i].reset();
	    addphage   = true;
	    addphage1  = true;
	    phageAdded = false;
	    bacteriaAdded=false;
	    bacteriaAdded2=false;


 	    bacteriaCollection.removeAllElements();
 	    phageCollection.removeAllElements();
	    
	    initCounter(); // initialize the count of bacteria and phage populations

	    updatePars();

	}
	return err;
    }


    public void successTerminate(){
	System.out.println("~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~");
	System.out.println("case Number: " + caseNumber);
	System.out.println("output written to: " + standardFilename);
	System.out.println("simulation successfully terminated");
	System.out.println("~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~");
	stop();
    }

    public void updatePars(){	
	if (caseNumber==8){
	    if (pars[1]<parsMax[1]) 
		pars[1]+=parsInc[1];
	    else {		
		if (pars[0]<parsMax[0]){
		    pars[1]=parsMin[1];
		    pars[0]+=parsInc[0];
		    parsMax[1]=pars[0];
		    //		    parsMax[1]=(int)(0.5*(double)(pars[0]+1.));
		    oW.insertLine(1);
		}
		else successTerminate();
	    }
	}

	if (caseNumber!=8){
	    if (nPars==1){
		if (pars[0]<parsMax[0]) 
		    pars[0]+=parsInc[0];
		else successTerminate();
	    }

	    if (nPars==2){
		if (pars[0]<parsMax[0]) 
		    pars[0]+=parsInc[0];
		else {
		    if (pars[1]<parsMax[1]){
			pars[0]=parsMin[0];
			pars[1]+=parsInc[1];
			oW.insertLine(1);
		    }
		    else successTerminate();
		}
	    }
	    
	    if (nPars==0){ successTerminate();}
	}
    }

    public int2D[] getNeighbors(int2D r){
	int2D[] neighbors=new int2D[Nneighbors];
	for (int k=0;k<Nneighbors;k++)
	    neighbors[k]=add2D(r,connections[k]);
	return neighbors;
    }
    
    public int[] getNeighbors(int i){
	int[] neighbors;
	int2D[] neighbors2D;
	int2D i2D=index2D(i);
	neighbors2D=getNeighbors(i2D);
	neighbors=new int[Nneighbors];
	for (int j=0;j<Nneighbors;j++){
	    neighbors[j]=index1D(neighbors2D[j]);
	}
	return neighbors;       
    }

    public int2D periodic(int2D r){
	int2D ret=new int2D();
	if (periodicBC){
	    ret.x=(r.x+L)%L;
	    ret.y=(r.y+L)%L;
	}
	else{
	    if (ret.x>=L) ret.x=-1;
	    if (ret.y>=L) ret.y=-1;
	}
	return ret;
    }

    public int2D add2D(int2D r1, int2D r2){
	int2D ret=new int2D();
	ret.x=r1.x+r2.x;
	ret.y=r1.y+r2.y;
	ret=periodic(ret);
	return ret;
    }

    public int2D mult2D(int2D r, int m){
	int2D ret=new int2D();
	ret.x=m*r.x;
	ret.y=m*r.y;
	return ret;
    }

    public int index1D(int x,int y){
	return x*L+y;
    }


    public int index1D(int2D r){
	return r.x*L+r.y;
    }


    public int2D index2D(int i){
	int2D ret=new int2D();
	ret.y=i%L;
	ret.x=(i-ret.y)/L;
	return ret;
    }


    public int2D index2D(int x, int y){
	int2D ret=new int2D();
	ret.x=x%L;
	ret.y=y%L;
	return ret;
    }

    
    BinarySearchTree tree;
    public void initialCondition(){
	double toIns;
	tree = new BinarySearchTree( );
	currentT=0.0;
	ctime=0;
	//	pars[0]=expressionAa;
	//	pars[1]=waitingTimes[0];

	if (host[1+L].p!=null) System.out.println("status 5 "+host[1+L].p.populationStatus);
		
	//printTimestamps();       

    }

    public Timestamp getCurrentTimestamp(){		
	return (Timestamp)(tree.findMin());
    }

    public void updateTimestamp(double t, Timestamp currentTS){	
	tree.removeMin();
	currentTS.t=t;
	tree.insert(currentTS);
    }

    public void removeTimestamp(){	
	tree.removeMin();
    }
        

    public double meanExpression(){
	double meanExp=0.;
	int nP=0;

	for (int c=0;c<L*L;c++){
	    if (host[c].p!=null) {

		meanExp+=host[c].p.expression;
		nP+=1;
	    }


	}		
	    System.out.println(meanExp/(double)nP);
	    return meanExp/(double)nP;
    }

    public double meanwT(){
	double[] meanwT=new double[NPHAGE];
	int[] nP=new int[NPHAGE];

	for (int np=0;np<NPHAGE;np++){
	    meanwT[np]=0;
	    nP[np]=0;
	    for (int c=0;c<L*L;c++){
		if (host[c].p!=null && host[c].p.phageID==(100+np)) {
		    
		    meanwT[np]+=host[c].p.waitingTime;
		    nP[np]+=1;
		}
		
		
	    }		
	    System.out.println("phage "+np+": " + meanwT[np]/(double)(nP[np]));
	}
	return meanwT[0]/(double)(nP[0]);
    }



    public void countPop(){
	
	for (int ind=0;ind<bacteriaCollection.size();ind++) nBact[ind]=0;
	for (int ind=0;ind<phageCollection.size();ind++)    nPhag[ind]=0;
	for (int ind=0;ind<phageCollection.size();ind++)    nPhagExp[ind]=0;
	for (int ind=0;ind<phageCollection.size();ind++)    nPhagwT[ind]=0;
	
	for (int c=0;c<L*L;c++){
	    if (host[c].b!=null) {
		//		System.out.println(bacteriaCollection.elementAt(0));
		for (int id=0;id<bacteriaCollection.size();id++)
		    if (host[c].b.bactID==bacteriaCollection.elementAt(id)) nBact[id]+=1;
		//		if (host[c].b.bactID==1) nBact[1]+=1;	     
	    }
	    if (host[c].p!=null) {
		for (int id=0;id<phageCollection.size();id++)
		    if (host[c].p.phageID==(100+id)) {nPhag[id]+=1;nPhagExp[id]+=host[c].p.expression;nPhagwT[id]+=host[c].p.waitingTime;}

	    }
	}		
    }

    public void printTimestamps(){	  
	Timestamp min=(Timestamp)(tree.findMin());
	Timestamp max=(Timestamp)(tree.findMax());
	if (min!=null && max!=null){
	    min.print();
	    max.print();
	}
	else System.out.println("no time step to print");
    }



    public void updateLatticeBacteria(int ID, double waitingTime){
	for (int i=0;i<L*L;i++){
	    if (host[i].b!=null && host[i].b.bactID==ID)
		host[i].b.waitingTime=waitingTime;
	}
    }


    public void addAllBacteria(){
	if (bacteriaAdded) bacteriaAdded2=true;
	bacteriaAdded=true;
	int bactType=0;
	Vector<Integer> imm0 = new Vector<Integer>();
	Vector<Integer> fam0 = new Vector<Integer>();
	
	if (caseNumber==1) {
	    bactType=0;
	    imm0.add(0); // adding an immunity to the bacterial vector	
	}
	if (caseNumber==2) {
	    bactType=0;
	    imm0.add(0); // adding an immunity to the bacterial vector	
	}
	if (caseNumber==3) {
	    bactType=2;
	    imm0.add(0); // adding an immunity to the bacterial vector
	    
	}
	if (caseNumber==4) bactType=3;

	if (caseNumber==5) {
	    bactType=0;
	    imm0.add(0); // adding an immunity to the bacterial vector	
	    //imm0.add(1);
	}
	if (caseNumber==6) {bactType=3;NPHAGE=(int)(pars[0]);}
	if (caseNumber==8) {bactType=4;NPHAGE=(int)(pars[0]);}
	if (caseNumber==7) bactType=3;

	if (caseNumber==9) bactType=5;


	if (pars!=null) System.out.println(pars[0]);
	for (int j=0;j<L*L;j+=(int)(L*Math.random())){
	    //	    System.out.println("adding");
	    if (caseNumber==1)                host[j].addBacteria(currentT, bactType, pars[0], pars[1], imm0, tree);
	    if (caseNumber==2)                host[j].addBacteria(currentT, bactType, pars[0], pars[1], imm0, tree);
	    if (caseNumber==3 && wTevolve && expevolve)    host[j].addBacteria(currentT, bactType, 0.0    ,     1.0, imm0, tree);
	    if (caseNumber==3 && (!wTevolve) && expevolve) host[j].addBacteria(currentT, bactType, 0.0    , pars[0], imm0, tree);
	    if (caseNumber==3 && wTevolve && (!expevolve)) host[j].addBacteria(currentT, bactType, 0.0    , pars[0], imm0, tree);
	    if (caseNumber==4)                host[j].addMuteBacteriaStandard(currentT, bactType, 0.0    , pars[0], imm0, tree,pars[1],NRES);
	    if (caseNumber==5)                host[j].addBacteria(currentT, bactType, pars[0], waitingTimes[2], imm0, tree);
	    if (caseNumber==6)                host[j].addMuteBacteriaStandard(currentT, bactType, 0.0    , 1.0, imm0, tree,RM,NRES);
	    if (caseNumber==8)                host[j].addMuteBacteria(currentT, bactType, 0.0    , 1.0, imm0, tree,RM,(int)(pars[1]));
	    if (caseNumber==7)                host[j].addMuteBacteria(currentT, bactType, 0.0    , 1.0, imm0, tree,pars[0],1);
	    //	    if (caseNumber==9)                host[j].addMute_nres_Bacteria(currentT, bactType, 0.0    , pars[0], imm0, tree,pars[1],NRES,immSwitchRate);
	    if (caseNumber==9 && zeroInitRes)                host[j].addMute_nres_Bacteria(currentT, bactType, 0.0    , pars[0], imm0, fam0, tree,pars[1],0,immSwitchRate,pars[0], rPrime, attackSurvival,cost);
	    if (caseNumber==9 && !zeroInitRes){
                int nr=(int)(Math.rint(Math.random()*NPHAGE));
		host[j].addMute_nres_Bacteria(currentT, bactType, 0.0    , pars[0], imm0, fam0, tree,pars[1],0,immSwitchRate,pars[0], rPrime, attackSurvival,cost);
		double rn=Math.random();
		if (rn>.5)
		    for (int i=0;i<nr;i++)   host[j].b.makeImmune(i,0.);	    
		if (rn<.5)
		    for (int i=0;i<nr;i++)   host[j].b.makeImmune(NPHAGE-1-i,0.);	    
	    }
	    if (!bacteriaCollection.contains(bactType)) bacteriaCollection.add(bactType);
	}
    }

    public void handleOutput(){
			    if (caseNumber==1) oW.outputToFile(pars,nBactTot,bacteriaCollection.size(),nPhagTot,phageCollection.size(),(double)(L*L));
			    if (caseNumber==2) oW.outputToFile(pars,nBactTot,bacteriaCollection.size(),nPhagTot,phageCollection.size(),(double)(L*L));
			    if (caseNumber==3){
				Histogram hist_expr=null, hist_wT=null;
				if (expevolve){
				    Vector<Double> numbers = new Vector<Double>();
				    for (int n=0;n<L*L;n++) {
					if (host[n].p!=null && host[n].p.expression>=0.0) {
					    numbers.addElement(host[n].p.expression);
					}
				    }				
				    hist_expr= new Histogram(numbers,20,0,1.00001,"expr");
				
				if (!wTevolve)
				    oW.outputToFile(pars,hist_expr,nBactTot,bacteriaCollection.size(),nPhagTot,phageCollection.size(),(double)(L*L));
				}
				if (wTevolve){
				    Vector<Double> numbers = new Vector<Double>();
				    for (int n=0;n<L*L;n++) {
					if (host[n].p!=null) {
					    numbers.addElement(host[n].p.waitingTime);
					}
				    }				
				    hist_wT  = new Histogram(numbers,30,0,2.00001,"wT");
				
				    if (!expevolve){
					oW.outputToFile(pars,hist_wT,nBactTot,bacteriaCollection.size(),nPhagTot,phageCollection.size(),(double)(L*L));			       
					oW_hist.outputToFile(pars,hist_wT);
				    }
				}
			
				if (wTevolve && expevolve)
				    oW.outputToFile(hist_expr,hist_wT,nBactTot,bacteriaCollection.size(),nPhagTot,phageCollection.size(),(double)(L*L));
			    
			    }
				
			    if (caseNumber==4) {
				if (!countExp) oW.outputToFile(pars,nBactTot,bacteriaCollection.size(),nPhagTot,phageCollection.size(),(double)(L*L));
				if (countExp) oW.outputToFile(pars,phagExp,phagAdd,nBactTot,bacteriaCollection.size(),nPhagTot,phageCollection.size(),(double)(L*L));
				
			    }
			    if (caseNumber==5) oW.outputToFile(pars,nBactTot,bacteriaCollection.size(),nPhagTot,phageCollection.size(),(double)(L*L));
			    if (caseNumber==6) oW.outputToFile(pars,nBactTot,bacteriaCollection.size(),nPhagTot,phageCollection.size(),(double)(L*L));			   
			    if (caseNumber==7) {
				Histogram hist_wT[]=new Histogram[NPHAGE];
				for (int np=0;np<NPHAGE;np++){
				    hist_wT[np]=null;
				    Vector<Double> numbers = new Vector<Double>();
				    for (int n=0;n<L*L;n++) {
					if (host[n].p!=null && host[n].p.phageID==(100+np)) {
					    numbers.addElement(host[n].p.waitingTime);
					}
				    }				
				    hist_wT[np]  = new Histogram(numbers,20,0,2.0,"wT_"+np);
				    
				    
				}
				oW.outputToFile(pars,hist_wT,nBactTot,bacteriaCollection.size(),nPhagTot,phageCollection.size(),(double)(L*L));	
			    }
			    if (caseNumber==8) {
				Histogram hist_wT[]=new Histogram[NPHAGE];
				for (int np=0;np<NPHAGE;np++){
				    hist_wT[np]=null;
				    Vector<Double> numbers = new Vector<Double>();
				    for (int n=0;n<L*L;n++) {
					if (host[n].p!=null && host[n].p.phageID==(100+np)) {
					    numbers.addElement(host[n].p.waitingTime);
					}
				    }				
				    hist_wT[np]  = new Histogram(numbers,20,0,2.00001,"wT_"+np);
				    oW_hist.outputToFile(pars,hist_wT[np]);
				}
				oW.outputToFile(pars,hist_wT,nBactTot,bacteriaCollection.size(),nPhagTot,phageCollection.size(),(double)(L*L));	
			    }
			    if (caseNumber==9) {
				Histogram hist_nRes;        
				int[][] neighborhood = new int[NPHAGE+4][NPHAGE+5];


				int nr, neigh;				
				for (int n=0;n<L*L;n++) {
				    if (host[n].b!=null) {
					nr=host[n].b.nRes;
					neighborhood[nr][NPHAGE+4]+=1;
					System.out.println("nei "+(host[n].neighborHosts).length);
					for (int i=0;i<(host[n].neighborHosts).length;i++) {
					    System.out.println(n+" "+i+" ");
					    System.out.println(host[n].neighborHosts[i].b!=null);
					    if (host[n].neighborHosts[i].b!=null) neigh=host[n].neighborHosts[i].b.nRes;
					    else if (host[n].neighborHosts[i].p!=null) neigh=NPHAGE+2;
					    else neigh=NPHAGE+3;
					    neighborhood[nr][neigh]+=1;

					}
				    }

				    else if (host[n].p!=null) {
					nr=NPHAGE+2;
					neighborhood[nr][NPHAGE+4]+=1;
					for (int i=0;i<(host[n].neighborHosts).length;i++) {
					    if (host[n].neighborHosts[i].b!=null) neigh=host[n].neighborHosts[i].b.nRes;
					    else if (host[n].neighborHosts[i].p!=null) neigh=NPHAGE+2;
					    else neigh=NPHAGE+3;
					    neighborhood[nr][neigh]+=1;	

					}
				    }
				    else { // host is empty site
					nr=NPHAGE+3;
					neighborhood[nr][NPHAGE+4]+=1;	    
					for (int i=0;i<(host[n].neighborHosts).length;i++) {
					    if (host[n].neighborHosts[i].b!=null) neigh=host[n].neighborHosts[i].b.nRes;
					    else if (host[n].neighborHosts[i].p!=null) neigh=NPHAGE+2;
					    else neigh=NPHAGE+3;
					    neighborhood[nr][neigh]+=1;	

					}
				    }
				}				

				oW_interaction.outputToFileIntMatrix2D(neighborhood,NPHAGE+4,NPHAGE+5,L*L);

				Vector<Double> nResvals = new Vector<Double>();
				Vector<Double> categories = new Vector<Double>();
				
				for (int c=0;c<=NPHAGE;c++) categories.addElement((double)c);


				for (int n=0;n<L*L;n++) {
				    if (host[n].b!=null) {
					nResvals.addElement((double)(host[n].b.nRes));
				    }
				}				
				hist_nRes = new Histogram(nResvals,categories,"nRes_");
				
				oW_hist.outputToFile(pars,hist_nRes);	


				Histogram hist_immSwi;        

				Vector<Double> immSwivals = new Vector<Double>();
				Vector<Double> Xvals = new Vector<Double>();
				Vector<Double> Yvals = new Vector<Double>();
				Vector<Vector<Integer>> families = new Vector<Vector<Integer>>();

				categories = new Vector<Double>();
				
				for (int c=0;c<=10;c++) {
				    categories.addElement(Math.pow(10.,rbMin)*Math.pow(2.,(double)c));
				    categories.addElement(Math.pow(10.,rbMin)*Math.pow(2.,-(double)c));
				}

				for (int n=0;n<L*L;n++) {
				    if (host[n].b!=null) {
					immSwivals.addElement((double)(host[n].b.immSwitchRate));
					Xvals.addElement((double)(host[n].x));
					Yvals.addElement((double)(host[n].y));
					families.addElement(host[n].b.family);
				    }
				}
				hist_immSwi = new Histogram(immSwivals,categories,"immSwi_");
				
				oW_hist1.outputToFile(pars,hist_immSwi);

				oW_scatter.outputToFileScatter4DFamily(Xvals,Yvals,nResvals,immSwivals,families);
				
				if (!countExp) oW.outputToFile(pars,nBactTot,bacteriaCollection.size(),nPhagTot,phageCollection.size(),(double)(L*L));
				if (countExp) oW.outputToFile(pars,phagExp,phagAdd,nBactTot,bacteriaCollection.size(),nPhagTot,phageCollection.size(),(double)(L*L));
				
			    }
    }

    public void resetMat(int[][] v, int nx, int ny){
	for (int i=0;i<nx;i++) for (int j=0;j<ny;j++) v[i][j]=0;
	//	System.out.println("te "+v[0][0]);
    }


    public void outputScatter(){
	
	Vector<Double> immSwivals = new Vector<Double>();
	Vector<Double> Xvals = new Vector<Double>();
	Vector<Double> Yvals = new Vector<Double>();
	Vector<Double> nResvals = new Vector<Double>();
	//Vector<Vector<Integer>> families = new Vector<Vector<Integer>>();
	
	for (int n=0;n<L*L;n++) {
	    if (host[n].b!=null) {
		immSwivals.addElement((double)(host[n].b.immSwitchRate));
		Xvals.addElement((double)(host[n].x));
		Yvals.addElement((double)(host[n].y));
		nResvals.addElement((double)(host[n].b.nRes));
		//		families.addElement(host[n].b.family);
	    }
	}
	oW_scatter_bact.insertComment("\n\n# xval   yval   rna ID");
	oW_scatter_bact.insertComment("# ctime = "+ctime);
	oW_scatter_bact.outputToFileScatter4D(Xvals,Yvals,nResvals,immSwivals);

	Xvals = new Vector<Double>();
	Yvals = new Vector<Double>();
	nResvals = new Vector<Double>();

	for (int n=0;n<L*L;n++) {
	    if (host[n].p!=null) {       
		nResvals.addElement((double)(host[n].p.rna));
		Xvals.addElement((double)(host[n].x));
		Yvals.addElement((double)(host[n].y));
	    }
	}
	oW_scatter_phage.insertComment("\n\n# xval   yval   rna ID");
	oW_scatter_phage.insertComment("# ctime = "+ctime);
	oW_scatter_phage.outputToFileScatter3D(Xvals,Yvals,nResvals);
    }

    public void outputTimeDept(){
	Histogram dum = bactCensus();
	Histogram dum1= phageCensus();
	for (int i=0;i<NPHAGE+1;i++) attackSurvival[i][5]=(int)Math.round((L*L*dum.density[i]));
	for (int i=0;i<NPHAGE+1;i++) attackSurvival[i][6]=(int)Math.round((L*L*dum1.density[i]));
	//System.out.println("attack survival "+attackSurvival[0][0]);
	oW_attackSurvival.insertComment("\n\n#phage: attack  survival  killed by other bact    killed other  total bactPop totalphage by Index");
	oW_attackSurvival.insertComment("# ctime = "+ctime);
	//System.out.println("le "+attackSurvival[0].length+"te "+attackSurvival[0][0]);
	oW_attackSurvival.outputToFileINTMatrix2D(attackSurvival,NPHAGE+1,attackSurvival[0].length);
	resetMat(attackSurvival,NPHAGE+1,attackSurvival[0].length);
    }

    public Histogram bactCensus(){
	Histogram hist_nRes;
	Vector<Double> categories = new Vector<Double>();
	Vector<Double> nResvals = new Vector<Double>();	
	for (int c=0;c<=NPHAGE;c++) categories.addElement((double)c);	       
	for (int n=0;n<L*L;n++) {
	    if (host[n].b!=null) {
		nResvals.addElement((double)(host[n].b.nRes));
	    }
	}				
	hist_nRes = new Histogram(nResvals,categories,"nRes_");	
	return hist_nRes;
    }

 public Histogram phageCensus(){
	Histogram hist_phage;
	Vector<Double> categories = new Vector<Double>();
	Vector<Double> indexVals = new Vector<Double>();	
	for (int c=0;c<=NPHAGE;c++) categories.addElement((double)c);	       
	for (int n=0;n<L*L;n++) {
	    if (host[n].p!=null) {
		indexVals.addElement((double)(host[n].p.rna));
	    }
	}				
	hist_phage = new Histogram(indexVals,categories,"nRes_");	
	return hist_phage;
    }

    public void addAllPhage(){ // we distinguish 4 cases
	phageAdded=true;
	if (addphage)  {
	    addphage=false;	
	    
	    if (caseNumber==1){ // standard case with 1 phage species
		for (int j=0;j<L*L;j+=(int)(L*Math.random()))
		    host[j].addPhage(currentT, waitingTimes[2], 100, tree);
		phageCollection.add(100);
	    }

	    if (caseNumber==2){ // as 1 but with 2 phage types
		for (int j=0;j<L*L;j+=(int)(L*Math.random()))
		    host[j].addPhage(currentT, waitingTimes[2], 100, tree);
		phageCollection.add(100);
		for (int j=0;j<L*L;j+=(int)(L*Math.random()))
		    host[j].addPhage(currentT, waitingTimes[2], 101, tree);
		phageCollection.add(101);
	    }

	    if (caseNumber==3){ // adding one evolving phage
		if (!wTevolve && expevolve){
		    System.out.println("adding exp evolving phage");
		    for (int j=0;j<L*L;j+=(int)(L*Math.random()))
			host[j].addMutePhage(currentT, waitingTimes[2], 100, tree, .7);//Math.random());
		    phageCollection.add(100);
		}
		if (wTevolve && expevolve){
		    System.out.println("adding totally evolving phage");
		    for (int j=0;j<L*L;j+=(int)(L*Math.random()))
			host[j].addMutePhage2(currentT, waitingTimes[2], 100, tree, .7);//Math.random());
		    phageCollection.add(100);
		}
		if (wTevolve && !expevolve){
		    System.out.println("adding wT-evolving phage");
		    for (int j=0;j<L*L;j+=(int)(L*Math.random()))
			host[j].addMutePhage3(currentT, waitingTimes[2], 100, tree, pars[0]);//Math.random());
		    phageCollection.add(100);
		}
		if (addExtraBacteria){
		    for (int j=0;j<L*L;j+=(int)(L*Math.random()))
		    host[j].addPhage(currentT, waitingTimes[2], 101, tree);
		    phageCollection.add(101);
		}

	    }
	    
	    if (caseNumber==4){ // two phage types and evolving bacteria
		for (int n=0;n<NPHAGE;n++){
		    for (int j=0;j<L*L;j+=(int)(L*Math.random())){
			if (!countExp) host[j].addPhage(currentT, waitingTimes[2], 100+n, tree);
			if (countExp) host[j].addPhageCountExp(currentT, waitingTimes[2], 100+n, tree, phagExp[n],phagAdd[n]); // this phage counts the number of successfull infections, otherwise it is the same as Phage.java
		    }
		    phageCollection.add(100+n);
		}
	    }
	    if (caseNumber==5){ // as 1 but with 2 phage types
		for (int j=0;j<L*L;j+=(int)(L*Math.random()))
		    host[j].addPhage(currentT, 1./pars[1], 100, tree);
		phageCollection.add(100);
		for (int j=0;j<L*L;j+=(int)(L*Math.random()))
		    host[j].addPhage(currentT, 1./2.2, 101, tree);
		phageCollection.add(101);
	    }
	    if (caseNumber==6){ // pars[0] phage types and evolving bacteria
		for (int n=0;n<pars[0];n++){
		    for (int j=0;j<L*L;j+=(int)(L*Math.random()))
			host[j].addPhage(currentT, waitingTimes[2], 100+n, tree);
		    phageCollection.add(100+n);
		}
	    }
	    if (caseNumber==8){ // pars[0] phage types and evolving bacteria
		for (int n=0;n<pars[0];n++){
		    for (int j=0;j<L*L;j+=(int)(L*Math.random()))
			host[j].addMutePhage3(currentT, waitingTimes[2]*2., 100+n, tree, 0.0);
		    phageCollection.add(100+n);
		}
	    }
	    if (caseNumber==7){ // two phage types and evolving bacteria
		for (int n=0;n<NPHAGE;n++){
		    for (int j=0;j<L*L;j+=(int)(L*Math.random()))
			host[j].addMutePhage3(currentT, waitingTimes[2], 100+n, tree, 0.0);
		    //			host[j].addPhage(currentT, waitingTimes[2], 100+n, tree);
		    phageCollection.add(100+n);
		}
	    }
	    if (caseNumber==9){ // two phage types and evolving bacteria
		System.out.println("adding phage");
		for (int n=0;n<NPHAGE;n++){
		    for (int j=0;j<L*L;j+=(int)((int)(L/5)*L*Math.random())){ // change back to L
			if (!countExp) host[j].addPhage(currentT, waitingTimes[2], 100+n, tree);
			if (countExp) host[j].addPhageCountExp(currentT, waitingTimes[2], 100+n, tree, phagExp[n],phagAdd[n]); // this phage counts the number of successfull infections, otherwise it is the same as Phage.java
		    }
		    phageCollection.add(100+n);
		}
	    }
	}
    }

    
    /**
     * At every update updateEveryM timesteps are taken, lattice is redrawn and plot is updated
     */
	public void update(Graphics g) {		
	Timestamp currentts;
	ctime+=1;        	
	for (int i = 0; i < this.updateEveryM; ++i) {	  
	    if (ctime>-1){
	    
	   ctime+=1;	   
	    
	 //if (ctime>2000000 && ctime<2000000+4*L*L/20.)
	 //  if (ctime%(L*L/20.)==0) printImage();

	    if (ctime<10 && !bacteriaAdded){
		addAllBacteria();
	    }
	    
	    if (ctime<(6*L*L) && !phageAdded){
		if (ctime%30000==0) {
		    countPop();
		    if ((double)((nBact[0])/(double)(L*L)) > .8) //change back to .5
			addAllPhage();
		}
	    }

	    

	    /*if (ctime<(15*L*L) && !bacteriaAdded2 && phageAdded){
		if (ctime%(14*L*L)==0) {
		    addAllBacteria();
		}
		}*/ //commented because it will look weird in time dependence
	   
	    if (ctime%(L*L)==0) outputTimeDept();
	    if (ctime%(10*L*L)==0) outputScatter(); // anything less than 10*L*L seems redundant (temporally correlated)
	    if (ctime>=(15*L*L)){ //change back to 15!!!
		if (ctime%(10*L*L)==0){//change back to 10!!!
		    countPop();
		    //outputTimeDept();
		    System.out.println("ctime "+ctime+"\n");
		    if (caseNumber==3 || caseNumber==7 || caseNumber==8) {
			meanExpression();
			meanwT();
		    }
		    
		    for (int b=0; b<bacteriaCollection.size(); b++) {
			nBactTot[b].addElement(nBact[b]);
			nBactTotLS[b].addy(nBact[b]);
			nBactTotLS[b].addx((double)ctime);
		    }
		    for (int p=0; p<phageCollection.size(); p++)    {
			nPhagTot[p].addElement(nPhag[p]);
			nPhagTotLS[p].addy(nPhag[p]);
			nPhagTotLS[p].addx((double)ctime);
			
			if (caseNumber==3){
			    if (expevolve){
				nPhagExpTot[p].addElement(nPhagExp[p]);		
				nPhagExpTotLS[p].addy(nPhag[p]);
				nPhagExpTotLS[p].addx((double)ctime);
			    }
			    if (wTevolve){
				nPhagwTTot[p].addElement(nPhagwT[p]);
				nPhagwTTotLS[p].addy(nPhag[p]);
				nPhagwTTotLS[p].addx((double)ctime);
			    }
			}
			if (caseNumber==8){
			    	nPhagwTTot[p].addElement(nPhagwT[p]);
				nPhagwTTotLS[p].addy(nPhag[p]);
				nPhagwTTotLS[p].addx((double)ctime);
			}
		    }
		    if (ctime%(20*L*L)==0){//change back to 100!!!			
			double ratiobact=0,ratiophag=0, ratiophagExp=0, ratiophagwT=0;
			for (int b=0; b<bacteriaCollection.size(); b++){
			    nBactTotLS[b].compute();
			    if (nBactTot[b].getCurr()==0.0) nBactTot[b].set_old_ave(0.0);
			    else nBactTot[b].set_old_ave(nBactTot[b].getAverage());				    

			    nBactTot[b].reset();
			    System.out.println("bact "+b+":\t "+nBactTotLS[b].getCoeffs()[0]+ "\t " +nBactTotLS[b].getCoeffs()[1]+ "\t " + nBactTot[b].get_old_ave()/(L*L));

			    if (Math.abs(nBactTotLS[b].getCoeffs()[1])<convCritSlope) nBactTotLS[b].setConverged(true);
			}

			for (int p=0; p<phageCollection.size(); p++){
			    if (countExp) {
				
				phagExp[p].set_old_ave(phagExp[p].getAverage());
				phagAdd[p].set_old_ave(phagAdd[p].getAverage());

				System.out.println("ExpLevel "+p+": "+phagExp[p].getOldAverage());
				System.out.println("AddLevel "+p+": "+phagAdd[p].getOldAverage());

				phagExp[p].reset();
				phagAdd[p].reset();
			    }
			    nPhagTotLS[p].compute();
			    if (nPhagTot[p].getCurr()==0.0) nPhagTot[p].set_old_ave(0.0);
			    else nPhagTot[p].set_old_ave(nPhagTot[p].getAverage());
			    
			    
			    if (caseNumber==3){
				if (expevolve){
				    nPhagExpTotLS[p].compute();
				    if (nPhagExpTot[p].getCurr()==0.0) nPhagExpTot[p].set_old_ave(0.0);
				    else nPhagExpTot[p].set_old_ave(nPhagExpTot[p].getAverage());	
				}
				if (wTevolve){
				    nPhagwTTotLS[p].compute();
				    if (nPhagwTTot[p].getCurr()==0.0) nPhagwTTot[p].set_old_ave(0.0);
				    else nPhagwTTot[p].set_old_ave(nPhagwTTot[p].getAverage());	
				}
			    }
			    if (caseNumber==8){
				nPhagwTTotLS[p].compute();
				if (nPhagwTTot[p].getCurr()==0.0) nPhagwTTot[p].set_old_ave(0.0);
				else nPhagwTTot[p].set_old_ave(nPhagwTTot[p].getAverage());	
			    }
			    				    			   			    
			    nPhagTot[p].reset();
			    nPhagExpTot[p].reset();
			    nPhagwTTot[p].reset();
			    System.out.println("phage "+p+":\t "+nPhagTotLS[p].getCoeffs()[0]+ "\t " +nPhagTotLS[p].getCoeffs()[1]+ "\t " + nPhagTot[p].get_old_ave()/(L*L));
			    if (caseNumber==3){
				if (expevolve)
				    System.out.println("Exp "+p+":\t "+nPhagExpTotLS[p].getCoeffs()[0]+ "\t " +nPhagExpTotLS[p].getCoeffs()[1]+ "\t " + nPhagExpTot[p].get_old_ave()/(L*L)); 
				if (wTevolve) System.out.println("wT "+p+":\t "+nPhagwTTotLS[p].getCoeffs()[0]+ "\t " +nPhagwTTotLS[p].getCoeffs()[1]+ "\t " + nPhagwTTot[p].get_old_ave()/(L*L)); 
			    }



			    if (Math.abs(nPhagTotLS[p].getCoeffs()[1])<convCritSlope) nPhagTotLS[p].setConverged(true);
			    if (caseNumber==3){				
				if (expevolve)
				    if (Math.abs(nPhagExpTotLS[p].getCoeffs()[1])<convCritSlope) nPhagExpTotLS[p].setConverged(true);
				if (wTevolve)
				    if (Math.abs(nPhagwTTotLS[p].getCoeffs()[1])<0.05*convCritSlope) nPhagwTTotLS[p].setConverged(true);
			    }
			    if (caseNumber==8){
				if (Math.abs(nPhagwTTotLS[p].getCoeffs()[1])<0.05*convCritSlope) nPhagwTTotLS[p].setConverged(true);
			    }
			}

			    
			
			if (nPars==2) System.out.println(pars[0] + "   " + pars[1] + "   " + (nBactTot[0].get_old_ave())/(double)(L*L) + " \nratiobact = " + ratiobact + "   " + nPhagTot[0].get_old_ave()/(double)(L*L) + " \nratiophag = " + ratiophag);	
			if (nPars==1) System.out.println(pars[0] + "   " + "   " + (nBactTot[0].get_old_ave())/(double)(L*L) + " \nratiobact = " + ratiobact + "   " + nPhagTot[0].get_old_ave()/(double)(L*L) + " \nratiophag = " + ratiophag + " \nratiophagExp = " + ratiophagExp);	
			if (nPars==0) System.out.println((nBactTot[0].get_old_ave())/(double)(L*L) + " ratiobact = " + ratiobact + "   " + nPhagTot[0].get_old_ave()/(double)(L*L) + " ratiophag = " + ratiophag+ " \nratiophagExp = " + ratiophagExp+ " \nratiophagwT = " + ratiophagwT);	
		    
			
			boolean conv=true;
			for (int b=0; b<bacteriaCollection.size(); b++) if (!nBactTotLS[b].getConverged()) conv=false;
			for (int b=0; b<phageCollection.size(); b++)    if (!nPhagTotLS[b].getConverged()) conv=false;
			if (caseNumber==3){
			    if (expevolve)
				for (int b=0; b<phageCollection.size(); b++)    if (!nPhagExpTotLS[b].getConverged()) conv=false;
			    if (wTevolve)
				for (int b=0; b<phageCollection.size(); b++)    if (!nPhagwTTotLS[b].getConverged()) conv=false;
			}
			if (caseNumber==8){
			    for (int b=0; b<phageCollection.size(); b++)    if (!nPhagwTTotLS[b].getConverged()) conv=false; 
			}
			////if (ctime>500000) conv=true;
			if (conv  || meanwT()>8.0 || nBactTot[0].get_old_ave()/(L*L)==1.) {
			    handleOutput();
			    resetLattice();
			    initialCondition();
			}
		    }
		}
	    }
		
	    currentts=getCurrentTimestamp();	    
	    
	    if (currentts!=null){		
		currentX=currentts.x;
		currentY=currentts.y;
		currentT=currentts.t;
		currentI=currentts.index;		
	    }	    	    	  	 
	    	      							       	       			
	    if (randomUpdate){// neighbors are randomly reassigned every time
		HostMute[] randomH=new HostMute[Nneighbors];
		for (int k=0;k<Nneighbors;k++) randomH[k]=host[(int)(Math.random()*L*L)];
		host[currentX * L + currentY].setNeighbors(randomH);
	    }
	    if (bitRandomUpdate){	       
		int m;
		HostMute[] h=new HostMute[Nneighbors];
		for (int k=0;k<Nneighbors;k++) {
		    m=1;
		    if (host[currentX * L + currentY].p!=null) m=1+(int)Math.rint(jumpPhage*Math.random());
		    if (host[currentX * L + currentY].b!=null) m=1+(int)Math.rint(jumpBact *Math.random());   
		    h[k]=host[index1D(add2D(index2D(currentX,currentY),mult2D(connections[k],m)))]; 
		}
		host[currentX * L + currentY].setNeighbors(h);
	    }
	    

	    host[currentX * L + currentY].update(currentI,currentT);
	    
	    }
	}
        //bufferedGraphics2D.setStroke(new BasicStroke(2));   
        redrawLattice(bufferedGraphics2D);
	
        //Graphics2D g2d = (Graphics2D) g;
        g.drawImage(background2D, 0, 0, this);
	
	}
    

    
    /**
     * Draws the lattice
     * @param g Graphics2D
     */
    public void redrawLattice(Graphics2D g) {
        int x, y;

        for (int i = 0; i < L; ++i) {
            for (int j = 0; j < L; ++j) {
                x = x0 + nodeSize * j;
                y = y0 + nodeSize * i;
                g.setColor(host[i * L + j].hostColor);
                g.fillRect(x, y, nodeSize, nodeSize);
            }
        }
    }

    
    /** 
     * Returns an ImageIcon, or null if the path was invalid. 
     */
    protected ImageIcon createImageIcon(String path) {
        java.net.URL imgURL = getClass().getResource(path);
        if (imgURL != null) {
            return new ImageIcon(imgURL);
        } else {
            System.err.println("Couldn't find file: " + path);
            return null;
        }
    }

    

    public void start() {
        animator = new Thread(this);
        animator.start();
        keepMoving = true;
    }


    public void stop() {
	animator = null;
	keepMoving = false;

    }
    
    public void printImage(){
	String imageString="noImage.dat";
	if (nPars==0) imageString="out/image"+"_ctime_"+ctime+standardFilename;
	if (nPars==1) imageString="out/image"+"pars0_"+pars[0]+"_ctime_"+ctime+standardFilename;
	if (nPars==2) imageString="out/image"+"pars0_"+pars[0]+"pars1_"+pars[1]+"_ctime_"+ctime+standardFilename;

	ImagePrinter iR=new ImagePrinter(imageString, true);
	iR.outputToFile(pars,L,host);
    }

    public void destroy() {
        removeAll();
    }

    public void run() {
        Thread thisThread = Thread.currentThread();
        while (animator == thisThread) {
            try {
                Thread.sleep(delay);
            } catch (InterruptedException e){}
            if(keepMoving) {
                repaint();
            }
        }
    }


    public void paint(Graphics g) {
        update(this.getGraphics());

    }

    public void mouseClicked(MouseEvent e) {
    }

    public void mousePressed(MouseEvent e) {
    }

    public void mouseReleased(MouseEvent e) {
    }

    public void mouseEntered(MouseEvent e) {
    }

    public void mouseExited(MouseEvent e) {
    }

    /**
     * Updates the shown value of alpha in applet.
     */
    public void updateAlphaText(Graphics2D g, int textOut_x, int textOut_y){
         g.setColor(cbackground);
         g.fillRect(textOut_x-5, textOut_y-10, 125, 80);// to cover text
         g.setColor(dgrey);
         g.setFont(new Font("Helvetica", Font.BOLD, 10));
         g.drawString("" +  String.format("%1$4.3f",  alphaN), textOut_x, textOut_y);

    }


    /**
     * Updates the shown value of generation time in applet.
     */
    public void updateGenTimes(Graphics2D g, int textOut_x, int textOut_y, double output){
         g.setColor(cbackground);
         g.fillRect(textOut_x-5, textOut_y-10, 125, 80);// to cover text
         g.setColor(dgrey);
         g.setFont(new Font("Helvetica", Font.BOLD, 10));
         g.drawString("" +  String.format("%1$4.3f",  output), textOut_x, textOut_y);

    }
    
    /**
     * Is activated if user adjusts alphavalue via scrollbar
     */
    public void adjustmentValueChanged(AdjustmentEvent e) {
        if (e.getSource() == scroll_alpha) {
            alphaN = alphaNMin * Math.pow(alphaNMax / alphaNMin, 1.0 * scroll_alpha.getValue() / nBINS);
            alpha = alphaN / (L * L);
            
            updateAlphaText(bufferedGraphics2D, alphaX, alphaY);
        }
	
	if (NBACT>0){
	    if (e.getSource() == scroll_genTime1) {
		waitingTimes[0]= (double)(scroll_genTime1.getValue()/100.);                       
		pars[1]=waitingTimes[0];
		updateLatticeBacteria(bacteriaCollection.elementAt(0),waitingTimes[0]);
		updateGenTimes(bufferedGraphics2D, gen1X, gen1Y,waitingTimes[0]);

	    }
	}
	
	if (NBACT>1){
	    if (e.getSource() == scroll_genTime2) {
		waitingTimes[1]= (double)(scroll_genTime2.getValue()/100.);                       		
		updateGenTimes(bufferedGraphics2D, gen2X, gen2Y,waitingTimes[1]);
	    }
	}
	
        if (e.getSource() == scroll_time) {
            updateEveryM = updateEveryM0 * scroll_time.getValue();
        }
        keepMoving = true;
    }
    
    
    /**
     * Inherited method that manages actions done by user.
     */
    public void actionPerformed(ActionEvent e) {
       
    	//Manage periodic button
        if (e.getSource() == button_periodicBC) {
           if(!periodicBC) {
	           periodicBC = true;
	           button_periodicBC.setLabel("Absorbing");
	           button_periodicBC.setBackground(orange);
           }
           else {
	           periodicBC = false;
	           button_periodicBC.setLabel("Periodic");
	           button_periodicBC.setBackground(lblue);
           }
        }
        
        //Manage stop button
        if (e.getSource() == button_stop) { 
            if (keepMoving) { //if 'stop' is pressed
                keepMoving = false; //stop updating
                animator = null;
                button_stop.setLabel("Run");
                button_stop.setBackground(lred);
            } else { //if 'run' is pressed
                keepMoving = true;
                animator = new Thread(this);
                animator.start();
                button_stop.setLabel("Pause");
            }
        }
        //Manage print button
        if (e.getSource() == button_print) { 
	    printImage();
        }
    }


} // end of class Phage



// just to helper class to produce averages
class makeAverage {
    double sum, old_ave,currElement;
    int nElements;
    boolean converged;
    String name;
    

    public makeAverage(){
	sum=0.;
	old_ave=-1.;
	nElements=0;
	converged=false;
	name=" ";
    }
    public void reset(){
	sum=0.;
	nElements=0;
    }
    public void setName(String s){
	this.name=s;
    }
    public String getName(){
	return name;
    }
    public void setConverged(boolean c){
	converged=true;
    }
    public boolean getConverged(){
	return converged;
    }
    public double getCurr(){
	return currElement;
    }
    public void addElement(double d){
	currElement=d;
	sum+=d;
	nElements+=1;
    }
    public void set_old_ave(double d){
	old_ave=d;
    }
    public double get_old_ave(){
	return old_ave;
    }
    public double getAverage(){
	return sum/(double)nElements;
    }	
    public double getOldAverage(){
	return old_ave;
    }
    public double giveRatio(){
	return getAverage()/old_ave;
    }
    public double giveNElements(){
	return nElements;
    }
}



class bactSwitcherMute_nres extends bactSwitcher{
    
    //    double immSwitchRate;
    double waitingTime0;
    double cost;
    //int[][] attackSurvival;
    //    Vector<Integer> family;

    public void init(int x, int y, double creationTime,int bactID, double expLevel, double waitingTime, int populationStatus, int maxImmunity, Vector<Integer> imm, Vector<Integer> fam, int Nneighbors, int2D[] neighbors, Host[] neighborHosts,BinarySearchTree t, double rb, int nRes,double immSwitchRate, double waitingTime0,double rPrime, int[][] attackSurvival,double cost){

	this.family=new Vector<Integer>(fam);
	this.immSwitchRate=immSwitchRate;
	this.rPrime=rPrime;
	this.attackSurvival=attackSurvival;
	//	if (immSwitchRate!=.001) System.out.println("test imm " + immSwitchRate);
	this.waitingTime0=waitingTime0;
	this.waitingTime=waitingTime;

	this.nRes=nRes;
	
	this.cost=cost;
	
	//	if (this.nRes>0) System.out.println("nres "+nRes);
	
	attackSurvival[this.nRes][4]+=1;  // recording growth into empty space
	
	double dummy=Math.random();
	if (dummy<rPrime) {
	    System.out.println("immSwiRt "+this.immSwitchRate);
	    if (dummy>(rPrime/2.)) this.immSwitchRate*=(Math.random()*2.);
	    else this.immSwitchRate/=(Math.random()*2.);
	    if (this.immSwitchRate>1.) this.immSwitchRate=1.;
	    
	    ////	    if (this.immSwitchRate>.002) System.out.println("immswitch "+this.immSwitchRate);
	}
	
	this.waitingTime=waitingTime0*(1.+this.cost*(double)(this.nRes));

	//	System.out.println("waiting Time "+this.waitingTime);

	if (dummy<immSwitchRate) {
	    
	    
	    this.nRes-=1;//Math.round(5.*Math.random());

	    if (this.nRes<0) this.nRes=0;
	    if (this.nRes>100) this.nRes=100;



	    ////if (false) if (Math.random()>.5) this.nRes+=1; else this.nRes-=1;
	    ////this.nRes+=(int)(Math.round(6.*(Math.random()-.5)));
	    this.waitingTime=waitingTime0*(1.+this.cost*(double)(this.nRes));
	    //	    System.out.println("waiting Time "+this.waitingTime);
	
	    //System.out.println("NRES "+this.nRes);
	}

	
	
	//	if (this.waitingTime>1.0) System.out.println("Waiting1 " + this.waitingTime + " NRES "+ this.nRes);

	init(x,y,creationTime, bactID, expLevel, this.waitingTime, 1,100,imm, Nneighbors,neighbors,neighborHosts,t,rb,this.nRes);
	adjustImmune();	    
    }

    public boolean init_Tent(int x, int y, double creationTime,int bactID, double expLevel, double waitingTime, int populationStatus, int maxImmunity, Vector<Integer> imm, Vector<Integer> fam, int Nneighbors, int2D[] neighbors, Host[] neighborHosts,BinarySearchTree t, double rb, int nRes,double immSwitchRate,Bacteria b, double waitingTime0, double rPrime, int[][] attackSurvival,double cost){

	boolean inserted=false;

	//	System.out.println("test");
	this.family=new Vector<Integer>(fam);
	this.t=t;
	this.rPrime=rPrime;
	this.waitingTime=waitingTime;
	this.attackSurvival=attackSurvival;
	this.waitingTime0=waitingTime0;
	this.ts=new Timestamp(x,y,getUpdateTime(creationTime),bactID);
	this.nRes=nRes;
	this.cost=cost;
	
	//	if (this.nRes>0) System.out.println("nres "+nRes);

	this.waitingTime=waitingTime0*(1.+this.cost*(double)(this.nRes));
	//	System.out.println("waiting Time "+this.waitingTime);
	
	//	if (immSwitchRate!=.001) System.out.println("test imm " + immSwitchRate);

	//System.out.println("getT "+(this.ts).getT()+ "  " + (b.ts).getT());
	//System.out.println("nres "+this.nRes+ "  " + b.nRes);
	////if (this.waitingTime>1.0 || b.waitingTime>1.0){
	    //    System.out.println("waitingTime "+this.waitingTime+ "  " + b.waitingTime + "\n");
	    //    System.out.println((this.ts).getT()+" "+(b.ts).getT());
	////	}

	////if ((this.ts).getT()<(b.ts).getT()) {
	    //    if (this.waitingTime>1.0 || b.waitingTime>1.0)
	    // System.out.println((this.ts).getT()+" "+(b.ts).getT());
	if (b.nRes!=this.nRes) {
	    attackSurvival[b.nRes][2]+=1; // this.b.nRes killed by dummy.nRes
	    attackSurvival[this.nRes][3]+=1;  // dummy.nRes killed this.b.nRes
	}
	    b.decrease(0);
	    inserted=true;
	    this.immSwitchRate=immSwitchRate;
	    double dummy=Math.random();
	    if (dummy<rPrime) {
		if (dummy>(rPrime/2.)) this.immSwitchRate*=(Math.random()*2.);
		else this.immSwitchRate/=(Math.random()*2.);
		if (this.immSwitchRate>1.) this.immSwitchRate=1.;
		////		if (this.immSwitchRate>.002) System.out.println("immswitch "+this.immSwitchRate);
	    }
    
	    if (dummy<immSwitchRate) {
	
		////if (false) if (Math.random()>.5) this.nRes+=1; else this.nRes-=1;
		this.nRes-=1;//Math.round(5.*Math.random());
		if (this.nRes<0) this.nRes=0;
		if (this.nRes>100) this.nRes=100;
					       		
		this.waitingTime=waitingTime0*(1.+this.cost*(double)(this.nRes));
		//		System.out.println("waiting Time "+this.waitingTime);
			
	    }
	    
	    //System.out.println("nres1 "+nRes);
	    //    adjustImmune();	    
	    //	    	if (this.waitingTime>1.0) System.out.println("Waiting2 " + this.waitingTime);
	    init(x,y,creationTime, bactID, expLevel, this.waitingTime, 1,100,imm, Nneighbors,neighbors,neighborHosts,t,rb,this.nRes);
	    //	    System.out.println("nres wT "+nRes+" "+waitingTime);
	    adjustImmune();	  
	    ////}
	return inserted;	
    }

   public void makeImmune(int rna) {		
       //if (false){ // remove this when error is found!!
	if (!immunity.contains(rna)){
	    if (Math.random()>immRate){
		attackSurvival[this.nRes][1]+=1; // record survival
		//System.out.println(attackSurvival[this.nRes][1]);
		family.add(rna+1);
		immunity.add(rna);
		this.nRes+=1;
	    }
	    else attackSurvival[this.nRes][0]+=1; // record death
	}
	else attackSurvival[this.nRes][1]+=1; // record survival
	//  }
   }

    public void makeImmune(int rna, double immRate) {		
       //if (false){ // remove this when error is found!!
	if (!immunity.contains(rna)){
	    if (Math.random()>immRate){
		attackSurvival[this.nRes][1]+=1; // record survival
		//System.out.println(attackSurvival[this.nRes][1]);
		family.add(rna+1);
		immunity.add(rna);
		this.nRes+=1;
	    }
	    else attackSurvival[this.nRes][0]+=1; // record death
	}
	else attackSurvival[this.nRes][1]+=1; // record survival
	//  }
   }

    public void adjustImmune() {
	//	System.out.println("nres "+nRes);
	while (immunity.size()>nRes) {
	    for (int i=nRes;i<immunity.size();i++){
		family.add((-1-immunity.elementAt(0)));
		immunity.removeElementAt(0);
	    }		    
	}
    }

    void spread(double time){ // spread to neighboring sites
	for (int i=0;i<Nneighbors;i++){
	    //	    if (this.immSwitchRate!=0.001)
		///		System.out.println("imm te "+this.immSwitchRate);
	    //	    neighborHosts[i].addMuteBacteriaStandard(time, bactID, expLevel, waitingTime, immunity,t,rb,nRes);
	    //	    System.out.println("Bspread ");
	    neighborHosts[i].addMute_nres_Bacteria(time, bactID, expLevel, waitingTime, immunity, family, t, rb, nRes,this.immSwitchRate,waitingTime0,this.rPrime, attackSurvival,this.cost);
	}
    }
}


class bactSwitcher extends bactA{

    int[][] attackSurvival;
    double immRate=1.0; // the rate of acquiring a new immunity
    double rb=1.0;

    // int nRes;

    public void init(int x, int y, double creationTime,int bactID, double expLevel, double waitingTime, int populationStatus, int maxImmunity, Vector<Integer> imm, int Nneighbors, int2D[] neighbors, Host[] neighborHosts,BinarySearchTree t, double rb, int nRes){//init the bacteria population with some initial density, set the immRate
	//	if (waitingTime>1.) System.out.println("WAITING "+waitingTime);
	    init(x,y,creationTime,bactID,expLevel,waitingTime,populationStatus,maxImmunity,imm,Nneighbors,neighbors,neighborHosts,t);
	    
	    this.rb=rb;
	    this.immRate=(1-Math.pow(10.,-rb));
	    this.nRes=nRes;
	    //if (nRes>0) System.out.println("Nres "+this.nRes+"\n");
    }
    

    boolean expressImmunity(Phage p){
	boolean ret=false;
	int ind;
	makeImmune(p);
	if (immunity.contains(p.rna)){// the phage rna is contained in the response system
	    if (p.expression>=0.0) ret=(Math.random()>p.expression); // partial immunity		   
	    else ret=true;	   
	}
	return ret;	    
    }

    boolean expressImmunity(int rna){
	boolean ret=false;
	int ind;
	makeImmune(rna);
	if (immunity.contains(rna))// the phage rna is contained in the response system
	    ret=true;
	return ret;	    
    }

    public void makeImmune(Phage p) {		
	if (!immunity.contains(p.rna)){
	    if (Math.random()>immRate){
		immunity.removeAllElements();
		immunity.add(p.rna);
	    }	    	    
	}    
    }

    public void makeImmune(int rna) {		
	if (!immunity.contains(rna)){
	    if (Math.random()>immRate){
		attackSurvival[this.nRes][1]+=1; // record survival
		System.out.println("Te");
		immunity.add(rna);
		this.nRes+=1;
		//// never removing, just adding; change to "true" for removal
		if (false){
		if (immunity.size()>nRes) {
		    for (int i=nRes;i<immunity.size();i++){
			immunity.removeElementAt(0);
		    }		    
		}
		}
		//		immunity.removeAllElements();
		//immunity.add(rna);
	    }
	    else attackSurvival[this.nRes][0]+=1; // record death of bacterium
	}
	else // immunity.contains(rna)
	    attackSurvival[this.nRes][1]+=1; // record survival of bacterium
    }


    public void adjustImmune() {
	//	System.out.println("nres "+nRes);
	while (immunity.size()>nRes) {
	    for (int i=nRes;i<immunity.size();i++){
		//family.add((-immunity.elementAt(0)));
		immunity.removeElementAt(0);
	    }		    
	}
    }

    void spread(double time){ // spread to neighboring sites
	for (int i=0;i<Nneighbors;i++){
	    neighborHosts[i].addMuteBacteriaStandard(time, bactID, expLevel, waitingTime, immunity,t,rb,nRes);
	}
    }
}

class bactSwitcherPlus extends bactSwitcher{

    //    int nRes;


    public void init(int x, int y, double creationTime,int bactID, double expLevel, double waitingTime, int populationStatus, int maxImmunity, Vector<Integer> imm, int Nneighbors, int2D[] neighbors, Host[] neighborHosts,BinarySearchTree t, double rb, int nRes){//init the bacteria population with some initial density, set the immRate
	    init(x,y,creationTime,bactID,expLevel,waitingTime,populationStatus,maxImmunity,imm,Nneighbors,neighbors,neighborHosts,t);
	    this.rb=rb;
	    this.immRate=(1-Math.pow(10.,-rb));
	    this.nRes=nRes;
    }
    
    
    public void makeImmune(Phage p) {		
	if (!immunity.contains(p.rna)){
	    if (Math.random()>immRate){
		//		immunity.removeAllElements();
		immunity.add(p.rna);
		if (immunity.size()>nRes) {
		    for (int i=nRes;i<immunity.size();i++){
			//System.out.print(immunity.firstElement()+"\t");
			immunity.removeElementAt(0);
		    }
		    //System.out.println("");
		    
		}
		//		System.out.println("imm size "+immunity.size());
	    }	    	    
	}    
	}

    void spread(double time){ // spread to neighboring sites
	for (int i=0;i<Nneighbors;i++){
	    neighborHosts[i].addMuteBacteria(time, bactID, expLevel, waitingTime, immunity,t,rb, nRes);
	}
    }


}

class bactMute extends bactA{
   

    boolean expressImmunity(Phage p){
	
	boolean ret=false;
	int ind;
	if (immunity.contains(p.rna)){
	    if (p.expression>=0.0) ret=(Math.random()>p.expression); // partial immunity		   
	    else ret=true;
	}
	return ret;	    
    }
}

class bactA extends Bacteria{
    public void update(double time){
	Timestamp currentTS=this.ts;
	t.removeMin();
	currentTS.t=getUpdateTime(time);
	t.insert(currentTS);
	///	currentTS.print();

	int popStat=this.populationStatus;
	if (popStat<maxPopulationStatus && popStat>0){
	    this.populationStatus+=1;
	    this.age+=1;	   
	}
	if (popStat==maxPopulationStatus){
	    spread(time);
	}
    }
    public void decrease(int rna){
	//	System.out.println("decrease ");
	this.populationStatus=0; 
	t.remove(ts);
    } 
    public void decreaseBact(){
	
	this.populationStatus=0; 
	t.remove(ts);
    }    
    boolean expressImmunity(int rna){
	boolean ret=false;
	int ind;
	if (immunity.contains(rna)){//perfect immunity
	    ret=(Math.random()>this.expLevel); // partial immunity		   
	}
	return ret;	    
    }
}

class bactB extends Bacteria{
    public void update(double time){
	
	if (this.populationStatus<maxPopulationStatus && this.populationStatus>0){
	    this.populationStatus+=1;
	    this.age+=1;
	}
	if (this.populationStatus==maxPopulationStatus){
	    spread(time);
	}
    }
    public void decrease(int rna){
	this.populationStatus=0;  // kill all of the bacteria	    	
    }
    boolean expressImmunity(int rna){
	boolean ret=false;
	int ind;	
	if (immunity.contains(rna)){
	    ret=(Math.random()>this.expLevel); // partial immunity
	    }
	return ret;	    
    }
}



class int2D {
    int x, y;
    public int2D(){
    }
    public void init(int x, int y){
	this.x=x;this.y=y;
    }
}





class GreaterThan implements Comparator {

    public int compare(Object arg0, Object arg1) {
	double y = ((Timestamp)arg0).t;
	double x = ((Timestamp)arg1).t;
	if(x > y) {
	    return -1;
	} else if(x == y) {
	    return 0;
	} else {
	    return 1;
	}
    }
}



// Basic node stored in unbalanced binary search trees
// Note that this class is not accessible outside
// of this package.

class BinaryNode {
    // Constructors
    BinaryNode( Comparable theElement ) {
        element = theElement;
        left = right = null;
    }
    
    // Friendly data; accessible by other package routines
    Comparable element;      // The data in the node
    BinaryNode left;         // Left child
    BinaryNode right;        // Right child
}

