// -----------------------------------------------------------------------------
// Template.java
// -----------------------------------------------------------------------------

/*
 * =============================================================================
 * Applet that simulates the interaction of bacteria and phage on a lattice
 * By Ala (plus Martin, Ludvig, Jeppe and Jan)
 * =============================================================================
 */


//import java.io.FileNotFoundException;
//import java.io.FileOutputStream;
//import java.io.PrintStream;
import java.io.*;
import java.awt.*;
import java.awt.event.*;
import java.net.*;
import java.util.*;
import java.util.Hashtable;
import java.util.Vector;
import java.util.Random;
import java.util.Date;
import javax.swing.*;




public class PhageMutate extends Panel implements ActionListener, AdjustmentListener, Runnable, MouseListener {

    private static final long serialVersionUID = 1L;
    
    //crucial parameters
    public double waitingTime0=2.1;  // the initial ratio of reproduction times
    public double expressionAa0=0.7; // the initial expression level
    public boolean randomUpdate=false;//should random updating of sites be performed?

    //Constants: Dynamics
    public static int L = 150; //Size of lattice
    public static int Nneighbors = 4;
    public static int NBACT = 1;
    public int2D[] connections;
    public double[] waitingTimes=new double[4];
    public double initialTime=0.0;
    public boolean addphage, addphage1;
    public int ctime=0;
    public int maxBacteriaSpecies=20;
    public int maxPhageSpecies=20;
    public double expressionAa=expressionAa0;
    //    public boolean insertPhage_a=true;
    //public boolean insertPhage_b=false;
    public boolean phageAdded=false;
    public PrintWriter pw;


    Host[] randomH=new Host[Nneighbors];

    int delay = 2; 
    int maxImmunity = 2000;
    public int updateEveryM0 = L * L / 100;
    public int updateEveryM = 2000*updateEveryM0; //How many updates before lattice is redrawn
    //    public int diseasePool = 10 * L;
    double alpha = 0.05 / (L * L); //how often new diseases appear
	
    //Constants: Heights, widths and sizes
    public static int nodeSize = 2; //length and height of one node
    public static int PlotWidth = 200, PlotHeight = 100; //height and width of diversity plot
    public static int x0 = 20, y0 = 20; //starting coordinates for lattice
    public static int panelX = x0 + L*nodeSize + 50, panelY = 40; //x and y coordinates for panel
    static final int appletHeight = L*nodeSize+2*x0;
    static final int appletWidth = L*nodeSize + PlotWidth + 200; //height and width of Applet
        
    //Initializing parameters
    boolean keepMoving = true;
    boolean periodicBC = true;
    public HostMute[] host;
    //public Host[] host;
    URL codeBase;
    outputWriter oW=new outputWriter("mutate_1_phaseDiag_L"+L+"1bmutate_phage_test.dat",true);
    outputWriter oW_hist=new outputWriter("mutate_1_hist_L"+L+"1bmutate_phage_test.dat",true);

    //currentPosition etc.
    int currentX,currentY;
    double currentT;
    int currentI;
    //number of bacteria and phage on lattice
    double[] nBact=new double[maxBacteriaSpecies];
    double[] nPhag=new double[maxPhageSpecies];

    makeAverage[] nBactTot=new makeAverage[maxBacteriaSpecies];
    makeAverage[] nPhagTot=new makeAverage[maxPhageSpecies];
    

    double countAve=0.;
    int nPars=2;
    double[] pars = new double[nPars];

	
    int countT=0;
    long ti0=0;
    long ti1=0;
    
    

    //graphics
    private volatile Thread animator = null;
    Image background2D;
    Graphics2D bufferedGraphics2D;
    RenderingHints hints;
    BasicStroke line_stroke;
    //    Colormap cmapTW;
    
    GradientPaint redtowhite, redtowhite1;
    Color lblue, white, yellow, lyellow, lbluish, navyblue, black, golden, nColor[], beige, orange;
    Color dgreen, lgreen, red, lred, dgrey, green, mblack, lgrey;
    Color cbackground, cborder;
    Color ctext, cbuttontext;
    
    //Main buttons
    Button button_stop,button_print;
    Button button_periodicBC;
    
    //Scrollbars
    Scrollbar scroll_alpha, scroll_genTime1, scroll_genTime2;
    
    double alphaN = alpha*(L*L), alphaNMin = 0.001, alphaNMax = 10;
    int nBINS = 100;
    int alphaX, alphaY;
    int gen1X, gen1Y, gen2X, gen2Y;
    Scrollbar scroll_time;

    // vectors to hold a collection of the introduced phage and bacteria
    Vector<Integer>    phageCollection    = new Vector<Integer>();
    Vector<Integer> bacteriaCollection    = new Vector<Integer>();

    /** The tree root. */
    protected BinaryNode root;
    


	BufferedWriter out;
	// set up a file to write output to
    /*
 
    */

    public static void main(String[] args) {
	Frame f = new Frame();
	f.addWindowListener(new java.awt.event.WindowAdapter() {
		public void windowClosing(java.awt.event.WindowEvent e) {
		    System.exit(0);
		};
	    });

	PhageMutate templ = new PhageMutate();
	templ.setSize(800,800); // same size as defined in the HTML APPLET
	f.add(templ);
	f.pack();
	templ.init();
	f.setSize(1200,900 + 20); // add 20, seems enough for the Frame title,
	//	f.show();
	f.setVisible(true);
    }

    /**
     * Initializes the applet by invoking createGUI
     */
  
	public void init() {
        try {
            SwingUtilities.invokeAndWait(new Runnable() {
		    public void run() { 
			//codeBase = getCodeBase();
			createGUI(); 
		    }
		});
        } catch (Exception e) {
            System.err.println("createGUI didn't successfully complete");
        }

	/*
	
	
	try {
	    out = new BufferedWriter(new FileWriter("outfilename"));
	    out.write("aString");
	    out.close();
	} catch (IOException e) {
	}
	*/
    }
    
    /**
     * makes graphical user interface, made by Ala
     */
    public void createGUI() {
        setSize(appletWidth, appletHeight);
	
        background2D = this.createImage(appletWidth, appletHeight);
        bufferedGraphics2D = (Graphics2D) background2D.getGraphics();
        start();

	/*
	try{
	    pw = new PrintWriter(new FileWriter("test.dat"));
	    //this is equal to:
	    //pw = new PrintWriter(new FileWriter(filename, false));
	    
	}
    
	catch (IOException e) {
	    System.out.println("couldn't open");
	    e.printStackTrace();
	}
	*/


        //Define colours
        white = new Color(255, 255, 255);
        lred = new Color(255, 153, 0);
        lblue = new Color(190, 203, 255);
        red = new Color(255, 42, 42);
        navyblue = new Color(0, 81, 186);
        black = new Color(0, 0, 0);
        orange = new Color(232, 170, 63);
        dgrey = new Color(51, 51, 51);
        lgrey = new Color(150, 150, 150);
        lgreen = new Color(0, 128, 0);
        cbackground = white;

        line_stroke = new BasicStroke(2.0f, BasicStroke.CAP_BUTT, BasicStroke.JOIN_BEVEL); //???

        hints = new RenderingHints(RenderingHints.KEY_ALPHA_INTERPOLATION, RenderingHints.VALUE_ALPHA_INTERPOLATION_SPEED);
        hints.put(RenderingHints.KEY_ANTIALIASING, RenderingHints.VALUE_ANTIALIAS_ON);
        hints.put(RenderingHints.KEY_COLOR_RENDERING, RenderingHints.VALUE_COLOR_RENDER_SPEED);
        hints.put(RenderingHints.KEY_DITHERING, RenderingHints.VALUE_DITHER_DISABLE);
        hints.put(RenderingHints.KEY_FRACTIONALMETRICS, RenderingHints.VALUE_FRACTIONALMETRICS_ON);
        hints.put(RenderingHints.KEY_ANTIALIASING, RenderingHints.VALUE_ANTIALIAS_ON);
        hints.put(RenderingHints.KEY_RENDERING, RenderingHints.VALUE_RENDER_SPEED);
        hints.put(RenderingHints.KEY_STROKE_CONTROL, RenderingHints.VALUE_STROKE_NORMALIZE);
        hints.put(RenderingHints.KEY_TEXT_ANTIALIASING, RenderingHints.VALUE_TEXT_ANTIALIAS_OFF);


        //Buttons
        alphaX = panelX + 40;
        alphaY = panelY-4;
	gen1X  = panelX + 160;
	gen1Y  = panelY+90-4;
	gen2X  = panelX + 160;
	gen2Y  = panelY+135-4;
        //Button parameters
        int butWidth = 60;
        int butHeight = 25;
       

	bufferedGraphics2D.setFont(new Font("SansSerif", Font.BOLD, 10));


        // Add scroll bar Update Speed
        panelY += 45;
        bufferedGraphics2D.drawString("Update Speed", panelX, panelY-4);
        scroll_time = new Scrollbar(Scrollbar.HORIZONTAL, 10, 3, 1, 103);
        scroll_time.addAdjustmentListener(this);
        scroll_time.setBounds(panelX, panelY, 100, 15);
        scroll_time.setBackground(Color.white);
        add(scroll_time);
        scroll_time.setValue(updateEveryM/updateEveryM0);
        updateEveryM = updateEveryM0 * scroll_time.getValue();
               
        // Add scroll bar Gen Times, scrollbar for generation time of bacteria species 1 relative to phage reproduction time.
	panelY += 45;
       // bufferedGraphics2D.setFont(new Font("SansSerif", Font.BOLD, 10));
	if (NBACT>0) {
	    double minval=0.02;
	    double maxval=50.0;
	    //	    bufferedGraphics2D.setColor(Color.red);
	     bufferedGraphics2D.drawString("rel. bacteria 1 gen. time", panelX , panelY-4);
	    bufferedGraphics2D.drawString(""+minval, panelX-25 , panelY+11);
	    bufferedGraphics2D.drawString(""+maxval, panelX+105 , panelY+11);

	    
	    scroll_genTime1 = new Scrollbar(Scrollbar.HORIZONTAL, (int)(minval*100.), 50, 2, (int)(maxval*100.+25.));
	    scroll_genTime1.addAdjustmentListener(this);
	    scroll_genTime1.setBounds(panelX, panelY, 100, 15);
	    scroll_genTime1.setBackground(Color.white);
	    
	    add(scroll_genTime1);
	    
	    int genTime  = (int)(100.*waitingTimes[0]);
	    scroll_genTime1.setValue(genTime);	
	    updateGenTimes(bufferedGraphics2D, gen1X, gen1Y,(double)genTime);
	}

	if (NBACT>1){ 
	    // Add scroll bar Gen Times, this is the scrollbar for generation time of bacteria species 2 relative to phage reproduction
	    panelY += 45;    
	    bufferedGraphics2D.drawString("rel. bacteria 2 gen. time", panelX , panelY-4);
	    
	    scroll_genTime2 = new Scrollbar(Scrollbar.HORIZONTAL, 1, 50, 2, 5025);
	    scroll_genTime2.addAdjustmentListener(this);
	    scroll_genTime2.setBounds(panelX, panelY, 100, 15);
	    scroll_genTime2.setBackground(Color.white);
	    
	    add(scroll_genTime2);
	    
	    int genTime  = (int)(100.*waitingTimes[1]);
	    scroll_genTime2.setValue(genTime);	
	    updateGenTimes(bufferedGraphics2D, gen2X, gen2Y,(double)genTime);
	}
	

        //Add stop button
        panelY += 45;
        button_stop = new Button("Pause1");
        button_stop.setBounds(panelX, panelY, butWidth, butHeight);
        button_stop.setBackground(orange);
        add(button_stop);
        button_stop.addActionListener(this);

        panelY += 4*butHeight;


        panelY += 45;
        button_print = new Button("Print");
        button_print.setBounds(panelX, panelY, butWidth, butHeight);
        button_print.setBackground(orange);
        add(button_print);
        button_print.addActionListener(this);

        panelY += 4*butHeight;

        setLayout(null);
        addMouseListener(this);        
	initCounter(); // initialize the count of bacteria and phage populations
        initLattice(bufferedGraphics2D);
	initialCondition();	
        bufferedGraphics2D.setBackground(black);
    }
    

    public void initCounter(){

	for (int b=0; b<maxBacteriaSpecies; b++) {nBactTot[b]=new makeAverage();nBactTot[b].setName("bact"+b);}
	for (int p=0; p<maxPhageSpecies; p++)    {nPhagTot[p]=new makeAverage();nPhagTot[p].setName("phag"+p);}
    }

    
    /**
     * Draws lattice with initial infection
     * @param g	Graphics2D
     */
       public void initLattice(Graphics2D g) {
	
	   host = new HostMute[L * L];
	   //host=new Host[L*L];
        int x = 0, y = 0;
	// good ranges : 1 phage species with imm=1.  [0 .. 2]
	//             : 1 phage species with imm=0.5 [0 .. 15]
	
	waitingTimes[0]=waitingTime0; // 1st bacteria vs. phage
	waitingTimes[1]=3.0; // 2nd bacteria vs. phage
	waitingTimes[2]=1.; // 1st phage
	waitingTimes[3]=1.; // 2nd phage


        if (NBACT>0) {	   
	    scroll_genTime1.setValue((int)(100.*waitingTimes[0]));	
	    updateGenTimes(bufferedGraphics2D, gen1X, gen1Y,waitingTimes[0]);
	}
	if (NBACT>1) {
	    scroll_genTime2.setValue((int)(100.*waitingTimes[1]));	
	    updateGenTimes(bufferedGraphics2D, gen2X, gen2Y,waitingTimes[1]);
	}



	addphage =true;
	addphage1=true;

	//	nspecies =0;
	//	nBspecies=bacteriaCollection.size();

	connections=new int2D[Nneighbors];
	connections[0]=new int2D();
	connections[0].init(1,0);
	connections[1]=new int2D();
	connections[1].init(0,1);
	connections[2]=new int2D();
	connections[2].init(-1,0);
	connections[3]=new int2D();
	connections[3].init(0,-1);
	


        for (int i = 0; i < L; ++i) {
            for (int j = 0; j < L; ++j) {
		int2D r=index2D(i,j);
		//System.out.println(r.x+ " "+r.y);
		int2D[] neighbors=getNeighbors(r);
	
		x = x0 + nodeSize * j;
                y = y0 + nodeSize * i;

		host[i * L + j] = new HostMute(); //create host
		//host[i * L + j] = new Host(); //create host
                host[i * L + j].hostID = i * L + j; //name host
                host[i * L + j].init(i, j, maxImmunity, Nneighbors, neighbors); //initialize host
                	     
		//    infectHost(i * L + j, 0); // initial infection
		g.setColor(host[i * L + j].hostColor);
                g.fillRect(x, y, nodeSize, nodeSize);
            }
        }
	for (int i = 0; i < L; ++i) {
            for (int j = 0; j < L; ++j) {
		Host[] h=new Host[Nneighbors];
		for (int k=0;k<Nneighbors;k++) h[k]=host[index1D(add2D(index2D(i,j),connections[k]))];
		host[i*L+j].setNeighbors(h);
	    }
	}
	
       }
    
    // resets the lattice by removing all bacteria and phage from a given site
    public boolean resetLattice(){
	boolean err=false;
	for (int i=0;i<L*L;i++) if (host[i]==null) err=true;
	if (err==false) {
	    for (int i=0;i<L*L;i++) host[i].reset();
	    addphage   = true;
	    addphage1  = true;
	    phageAdded = false;


 	    bacteriaCollection.removeAllElements();
 	    phageCollection.removeAllElements();
	    
	    initCounter(); // initialize the count of bacteria and phage populations

	    updatePars();

	}
	return err;
    }


    public void updatePars(){
	
	//if (expressionAa>=1.0) {
	//  expressionAa=expressionAa0;
	    waitingTimes[0]+=.1; // 1st bacteria vs. phage
	    //  oW.insertLine(1);
	    //}

	    //else expressionAa+=0.025;
		
	pars[0]=expressionAa;
	pars[1]=waitingTimes[0];

    }

    public int2D[] getNeighbors(int2D r){
	int2D[] neighbors=new int2D[Nneighbors];
	for (int k=0;k<Nneighbors;k++)
	    neighbors[k]=add2D(r,connections[k]);
	return neighbors;
    }
    
    public int[] getNeighbors(int i){
	int[] neighbors;
	int2D[] neighbors2D;
	int2D i2D=index2D(i);
	neighbors2D=getNeighbors(i2D);
	neighbors=new int[Nneighbors];
	for (int j=0;j<Nneighbors;j++){
	    neighbors[j]=index1D(neighbors2D[j]);
	}
	return neighbors;       
    }

    public int2D periodic(int2D r){
	int2D ret=new int2D();
	if (periodicBC){
	    ret.x=(r.x+L)%L;
	    ret.y=(r.y+L)%L;
	}
	else{
	    if (ret.x>=L) ret.x=-1;
	    if (ret.y>=L) ret.y=-1;
	}
	return ret;
    }

    public int2D add2D(int2D r1, int2D r2){
	int2D ret=new int2D();
	ret.x=r1.x+r2.x;
	ret.y=r1.y+r2.y;
	ret=periodic(ret);
	return ret;
    }

    public int index1D(int x,int y){
	return x*L+y;
    }


    public int index1D(int2D r){
	return r.x*L+r.y;
    }


    public int2D index2D(int i){
	int2D ret=new int2D();
	ret.y=i%L;
	ret.x=(i-ret.y)/L;
	return ret;
    }


    public int2D index2D(int x, int y){
	int2D ret=new int2D();
	ret.x=x%L;
	ret.y=y%L;
	return ret;
    }

    
    BinarySearchTree tree;
    public void initialCondition(){
	double toIns;
	tree = new BinarySearchTree( );
	currentT=0.0;
	ctime=0;
	pars[0]=expressionAa;
	pars[1]=waitingTimes[0];

	Vector<Integer> imm0 = new Vector<Integer>();
	Vector<Integer> imm1 = new Vector<Integer>();	

	imm0.add(0); // adding an immunity to the bacterial vector

	for (int j=0;j<L*L;j+=(int)(L*Math.random())){
	    host[j].addBacteria(currentT, 2,expressionAa, waitingTimes[0], imm0, tree);// add some bacteria to a given site
	}
	    //	host[L*L/2+L/2].addBacteria(currentT, 2,expressionAa, waitingTimes[0], imm0, tree);// add some bacteria to a given site
	    //	host[L*L/2+L/2+1].addBacteria(currentT, 0,expressionAa, waitingTimes[0], imm0, tree);// add some bacteria to a given site
	//	}
	bacteriaCollection.add(2);


	if (host[1+L].p!=null) System.out.println("status 5 "+host[1+L].p.populationStatus);
		
	printTimestamps();       

    }

    public Timestamp getCurrentTimestamp(){		
	return (Timestamp)(tree.findMin());
    }

    public void updateTimestamp(double t, Timestamp currentTS){	
	tree.removeMin();
	currentTS.t=t;
	tree.insert(currentTS);
    }

    public void removeTimestamp(){	
	tree.removeMin();
    }
        

    public void meanExpression(){
	double meanExp=0.;
	int nP=0;

	for (int c=0;c<L*L;c++){
	    if (host[c].p!=null) {

		meanExp+=host[c].p.expression;
		nP+=1;
	    }


	}		
	    System.out.println(meanExp/(double)nP);
    }

    public void countPop(){
	
	for (int ind=0;ind<bacteriaCollection.size();ind++) nBact[ind]=0;
	for (int ind=0;ind<phageCollection.size();ind++)    nPhag[ind]=0;
	for (int c=0;c<L*L;c++){
	    if (host[c].b!=null) {
		//		System.out.println(bacteriaCollection.elementAt(0));
		for (int id=0;id<bacteriaCollection.size();id++)
		    if (host[c].b.bactID==bacteriaCollection.elementAt(id)) nBact[id]+=1;
		//		if (host[c].b.bactID==1) nBact[1]+=1;	     
	    }
	    if (host[c].p!=null) {
		for (int id=0;id<phageCollection.size();id++)
		    if (host[c].p.phageID==(100+id)) nPhag[id]+=1;
		//		if (host[c].p.phageID==3) nPhag[1]+=1;	     
	    }
	}		
    }

    public void printTimestamps(){	  
	Timestamp min=(Timestamp)(tree.findMin());
	Timestamp max=(Timestamp)(tree.findMax());
	min.print();
	max.print();
    }



    public void updateLatticeBacteria(int ID, double waitingTime){
	for (int i=0;i<L*L;i++){
	    if (host[i].b!=null && host[i].b.bactID==ID)
		host[i].b.waitingTime=waitingTime;
	}
    }

    public void addAllPhage(){
	//	System.out.println("adding phage");
	phageAdded=true;
	if (addphage)  {
	    addphage=false;	
	   
	    /*
	    double expLev=0.9;
	    for (int j=0;j<L*L;j+=(int)(L*Math.random()))
		host[j].addMutePhage(currentT, waitingTimes[2], 100, tree, Math.random());//expLev		
	    */
		
	  
	    for (int j=0;j<L*L;j+=(int)(L*Math.random()))
		host[j].addPhage(currentT, waitingTimes[2], 100, tree);//expLev		
	    
	    phageCollection.add(100);
	    
	    for (int j=1;j<L*L;j+=(int)(L*Math.random()))
		host[j].addPhage(currentT, waitingTimes[2], 101, tree);		
	    phageCollection.add(101);
	    
	    /*  
	      for (int j=1;j<L*L;j+=(int)(L*Math.random()))
	      host[j].addPhage(currentT, waitingTimes[2], 102, tree);		
	      phageCollection.add(102);
	      
	      for (int j=1;j<L*L;j+=(int)(L*Math.random()))
	      host[j].addPhage(currentT, waitingTimes[2], 103, tree);		
	      phageCollection.add(103);
	    */
	}
    }

    
    /**
     * At every update updateEveryM timesteps are taken, lattice is redrawn and plot is updated
     */
	public void update(Graphics g) {		
	Timestamp currentts;
	ctime+=1;        	
	for (int i = 0; i < this.updateEveryM; ++i) {	  
	if (ctime>-1){
	    ctime+=1;	   
	    
	    if (ctime<(6*L*L) && !phageAdded){
		if (ctime%30000==0) {
		    countPop();
		    if ((double)((nBact[0])/(double)(L*L)) > .5) 
			addAllPhage();
		}
	    }

	    //if (ctime<30) printImage();
	    
	    if (ctime>=(6*L*L)){
		if (ctime%(10*L*L)==0){
		    countPop();
		    meanExpression();
		    
		    for (int b=0; b<bacteriaCollection.size(); b++) {nBactTot[b].addElement(nBact[b]);}
		    for (int p=0; p<phageCollection.size(); p++)    {nPhagTot[p].addElement(nPhag[p]);}
		    if (ctime%(100*L*L)==0){
			double ratiobact=0,ratiophag=0;
			for (int b=0; b<bacteriaCollection.size(); b++){
			    if (nBactTot[b].getOldAverage()!=0.) ratiobact=nBactTot[b].getAverage()/nBactTot[b].getOldAverage();
			    else ratiobact=1.0;
				
			    nBactTot[b].set_old_ave(nBactTot[b].getAverage());	
			    
			    nBactTot[b].reset();
			    if (Math.abs(ratiobact-1.)<0.001) nBactTot[b].setConverged(true);
			    	    
			}
			for (int p=0; p<phageCollection.size(); p++){
			    if (nPhagTot[p].getOldAverage()!=0.) ratiophag=nPhagTot[p].getAverage()/nPhagTot[p].getOldAverage();
			    else ratiophag=1.0;

			    nPhagTot[p].set_old_ave(nPhagTot[p].getAverage());	

			    nPhagTot[p].reset();
			    if (Math.abs(ratiophag-1.)<0.001) nPhagTot[p].setConverged(true);

			}		    
		    
			System.out.println(pars[0] + "   " + pars[1] + "   " + (nBactTot[0].get_old_ave())/(double)(L*L) + " ratiobact = " + ratiobact + "   " + nPhagTot[0].get_old_ave()/(double)(L*L) + " ratiophag = " + ratiophag);	
		    
			
			boolean conv=true;
			for (int b=0; b<bacteriaCollection.size(); b++) if (!nBactTot[b].getConverged()) conv=false;
			for (int b=0; b<phageCollection.size(); b++)    if (!nPhagTot[b].getConverged()) conv=false;
			
			if (conv) {

			    Vector<Double> numbers = new Vector<Double>();
			    for (int n=0;n<L*L;n++) {
				if (host[n].p!=null) {
				    numbers.addElement(host[n].p.expression);
				}
			    }
			    Histogram hist=new Histogram(numbers,20,0,1.00001);
			    pars[0]=hist.mean;
			    oW.outputToFile(pars,nBactTot,bacteriaCollection.size(),nPhagTot,phageCollection.size(),(double)(L*L));
			    oW_hist.outputToFile(pars,hist);
			    resetLattice();
			    initialCondition();
			}
		    }
		}
	    }
		
	    currentts=getCurrentTimestamp();	    
	    
	    
	    //	    currentts.print();
	    if (currentts!=null){
		
		currentX=currentts.x;
		currentY=currentts.y;
		currentT=currentts.t;
		currentI=currentts.index;
		
	    }	    	    	  	 
		
	      							       	       			
	    if (randomUpdate){// neighbors are randomly reassigned every time
		for (int k=0;k<Nneighbors;k++) randomH[k]=host[(int)(Math.random()*L*L)];
		host[currentX * L + currentY].setNeighbors(randomH);
	    }

	    host[currentX * L + currentY].update(currentI,currentT);
	    
	    
	}
	}
        //bufferedGraphics2D.setStroke(new BasicStroke(2));   
        redrawLattice(bufferedGraphics2D);
	
        //Graphics2D g2d = (Graphics2D) g;
        g.drawImage(background2D, 0, 0, this);

	
	}
    

    
    /**
     * Draws the lattice
     * @param g Graphics2D
     */
    public void redrawLattice(Graphics2D g) {
        int x, y;

        for (int i = 0; i < L; ++i) {
            for (int j = 0; j < L; ++j) {
                x = x0 + nodeSize * j;
                y = y0 + nodeSize * i;
                g.setColor(host[i * L + j].hostColor);
                g.fillRect(x, y, nodeSize, nodeSize);
            }
        }
    }

    
    /** 
     * Returns an ImageIcon, or null if the path was invalid. 
     */
    protected ImageIcon createImageIcon(String path) {
        java.net.URL imgURL = getClass().getResource(path);
        if (imgURL != null) {
            return new ImageIcon(imgURL);
        } else {
            System.err.println("Couldn't find file: " + path);
            return null;
        }
    }

    

    public void start() {
        animator = new Thread(this);
        animator.start();
        keepMoving = true;
    }


    public void stop() {
	animator = null;
	keepMoving = false;

    }
    
    public void printImage(){
	ImagePrinter iR=new ImagePrinter("out/image_L"+L+"nB"+bacteriaCollection.size()+"+nP"+phageCollection.size()+"_ex_"+expressionAa+"_t0_"+(double)(((int)(1000.*waitingTimes[0]))/1000.)+"_ctime_"+ctime+".dat", true);
	iR.outputToFile(pars,L,host);
    }

    public void destroy() {
        removeAll();
    }

    public void run() {
        Thread thisThread = Thread.currentThread();
        while (animator == thisThread) {
            try {
                Thread.sleep(delay);
            } catch (InterruptedException e){}
            if(keepMoving) {
                repaint();
            }
        }
    }


    public void paint(Graphics g) {
        update(this.getGraphics());

    }

    public void mouseClicked(MouseEvent e) {
    }

    public void mousePressed(MouseEvent e) {
    }

    public void mouseReleased(MouseEvent e) {
    }

    public void mouseEntered(MouseEvent e) {
    }

    public void mouseExited(MouseEvent e) {
    }

    /**
     * Updates the shown value of alpha in applet.
     */
    public void updateAlphaText(Graphics2D g, int textOut_x, int textOut_y){
         g.setColor(cbackground);
         g.fillRect(textOut_x-5, textOut_y-10, 125, 80);// to cover text
         g.setColor(dgrey);
         g.setFont(new Font("Helvetica", Font.BOLD, 10));
         g.drawString("" +  String.format("%1$4.3f",  alphaN), textOut_x, textOut_y);

    }


    /**
     * Updates the shown value of generation time in applet.
     */
    public void updateGenTimes(Graphics2D g, int textOut_x, int textOut_y, double output){
         g.setColor(cbackground);
         g.fillRect(textOut_x-5, textOut_y-10, 125, 80);// to cover text
         g.setColor(dgrey);
         g.setFont(new Font("Helvetica", Font.BOLD, 10));
         g.drawString("" +  String.format("%1$4.3f",  output), textOut_x, textOut_y);

    }
    
    /**
     * Is activated if user adjusts alphavalue via scrollbar
     */
    public void adjustmentValueChanged(AdjustmentEvent e) {
        if (e.getSource() == scroll_alpha) {
            alphaN = alphaNMin * Math.pow(alphaNMax / alphaNMin, 1.0 * scroll_alpha.getValue() / nBINS);
            alpha = alphaN / (L * L);
            
            updateAlphaText(bufferedGraphics2D, alphaX, alphaY);
        }
	
	if (NBACT>0){
	    if (e.getSource() == scroll_genTime1) {
		waitingTimes[0]= (double)(scroll_genTime1.getValue()/100.);                       
		updateLatticeBacteria(0,waitingTimes[0]);
		updateGenTimes(bufferedGraphics2D, gen1X, gen1Y,waitingTimes[0]);

	    }
	}
	
	if (NBACT>1){
	    if (e.getSource() == scroll_genTime2) {
		waitingTimes[1]= (double)(scroll_genTime2.getValue()/100.);                       
		updateGenTimes(bufferedGraphics2D, gen2X, gen2Y,waitingTimes[1]);
	    }
	}
	
        if (e.getSource() == scroll_time) {
            updateEveryM = updateEveryM0 * scroll_time.getValue();
        }
        keepMoving = true;
    }
    
    
    /**
     * Inherited method that manages actions done by user.
     */
    public void actionPerformed(ActionEvent e) {
       
    	//Manage periodic button
        if (e.getSource() == button_periodicBC) {
           if(!periodicBC) {
	           periodicBC = true;
	           button_periodicBC.setLabel("Absorbing");
	           button_periodicBC.setBackground(orange);
           }
           else {
	           periodicBC = false;
	           button_periodicBC.setLabel("Periodic");
	           button_periodicBC.setBackground(lblue);
           }
        }
        
        //Manage stop button
        if (e.getSource() == button_stop) { 
            if (keepMoving) { //if 'stop' is pressed
                keepMoving = false; //stop updating
                animator = null;
                button_stop.setLabel("Run");
                button_stop.setBackground(lred);
            } else { //if 'run' is pressed
                keepMoving = true;
                animator = new Thread(this);
                animator.start();
                button_stop.setLabel("Pause");
            }
        }
        //Manage print button
        if (e.getSource() == button_print) { 
	    printImage();
        }
    }


} // end of class Phage



// just to helper class to produce averages
class makeAverage {
    double sum, old_ave,currElement;
    int nElements;
    boolean converged;
    String name;
    

    public makeAverage(){
	sum=0.;
	old_ave=-1.;
	nElements=0;
	converged=false;
	name=" ";
    }
    public void reset(){
	sum=0.;
	nElements=0;
    }
    public void setName(String s){
	this.name=s;
    }
    public String getName(){
	return name;
    }
    public void setConverged(boolean c){
	converged=true;
    }
    public boolean getConverged(){
	return converged;
    }
    public double getCurr(){
	return currElement;
    }
    public void addElement(double d){
	currElement=d;
	sum+=d;
	nElements+=1;
    }
    public void set_old_ave(double d){
	old_ave=d;
    }
    public double get_old_ave(){
	return old_ave;
    }
    public double getAverage(){
	return sum/(double)nElements;
    }	
    public double getOldAverage(){
	return old_ave;
    }
    public double giveRatio(){
	return getAverage()/old_ave;
    }
    public double giveNElements(){
	return nElements;
    }
}



class bactMute extends bactA{
   
    boolean expressImmunity(Phage p){
	
	boolean ret=false;
	int ind;
	if (immunity.contains(p.rna)){//perfect immunity
	    if (p.expression>=0.0) ret=(Math.random()>p.expression); // partial immunity		   
	    else ret=true;
	}
	return ret;	    
    }
}

class bactA extends Bacteria{
    public void update(double time){
	Timestamp currentTS=this.ts;
	t.removeMin();
	currentTS.t=getUpdateTime(time);
	t.insert(currentTS);
	///	currentTS.print();

	int popStat=this.populationStatus;
	if (popStat<maxPopulationStatus && popStat>0){
	    this.populationStatus+=1;
	    this.age+=1;	   
	}
	if (popStat==maxPopulationStatus){
	    spread(time);
	}
    }
    public void decrease(int rna){
	this.populationStatus=0; 
	t.remove(ts);
    }    
    boolean expressImmunity(int rna){
	boolean ret=false;
	int ind;
	if (immunity.contains(rna)){//perfect immunity
	    ret=(Math.random()>this.expLevel); // partial immunity		   
	}
	return ret;	    
    }
}

class bactB extends Bacteria{
    public void update(double time){
	
	if (this.populationStatus<maxPopulationStatus && this.populationStatus>0){
	    this.populationStatus+=1;
	    this.age+=1;
	}
	if (this.populationStatus==maxPopulationStatus){
	    spread(time);
	}
    }
    public void decrease(int rna){
	this.populationStatus=0;  // kill all of the bacteria	    	
    }
    boolean expressImmunity(int rna){
	boolean ret=false;
	int ind;	
	if (immunity.contains(rna)){
	    ret=(Math.random()>this.expLevel); // partial immunity
	    }
	return ret;	    
    }
}



class int2D {
    int x, y;
    public int2D(){
    }
    public void init(int x, int y){
	this.x=x;this.y=y;
    }
}





class GreaterThan implements Comparator {

    public int compare(Object arg0, Object arg1) {
	double y = ((Timestamp)arg0).t;
	double x = ((Timestamp)arg1).t;
	if(x > y) {
	    return -1;
	} else if(x == y) {
	    return 0;
	} else {
	    return 1;
	}
    }
}



// Basic node stored in unbalanced binary search trees
// Note that this class is not accessible outside
// of this package.

class BinaryNode {
    // Constructors
    BinaryNode( Comparable theElement ) {
        element = theElement;
        left = right = null;
    }
    
    // Friendly data; accessible by other package routines
    Comparable element;      // The data in the node
    BinaryNode left;         // Left child
    BinaryNode right;        // Right child
}

